from .residue import Residue, res_type

import mrich as logger


def alphabet(name):
    try:
        return _letter_from_resname[name]
    except KeyError:
        mout.errorOut(f"Unsupported amino acid residue name: {name}")
        return None


def longname(name):
    try:
        return _longname_from_resname[name]
    except KeyError:
        mout.errorOut(f"Unsupported amino acid residue name: {name}")
        return None


_letter_from_resname = {
    "ALA": "A",
    "ARG": "R",
    "ASN": "N",
    "ASP": "D",
    "CYS": "C",
    "GLN": "Q",
    "GLU": "E",
    "GLY": "G",
    "HIS": "H",
    "HSD": "H",
    "HSE": "H",
    "HSP": "H",
    "ILE": "I",
    "LEU": "L",
    "LYS": "K",
    "MET": "M",
    "PHE": "F",
    "PRO": "P",
    "SER": "S",
    "THR": "T",
    "TRP": "W",
    "TYR": "Y",
    "VAL": "V",
    "SEC": "U",
    "PYL": "O",
}

_resname_from_letter = {
    "A": "ALA",
    "R": "ARG",
    "N": "ASN",
    "D": "ASP",
    "C": "CYS",
    "Q": "GLN",
    "E": "GLU",
    "G": "GLY",
    "H": "HIS",
    "H": "HSD",
    "H": "HSE",
    "H": "HSP",
    "I": "ILE",
    "L": "LEU",
    "K": "LYS",
    "M": "MET",
    "F": "PHE",
    "P": "PRO",
    "S": "SER",
    "T": "THR",
    "W": "TRP",
    "Y": "TYR",
    "V": "VAL",
    "U": "SEC",
    "O": "PYL",
}

_longname_from_resname = {
    "ALA": "Alanine",
    "ARG": "Arginine",
    "ASN": "Asparagine",
    "ASP": "Aspartic Acid",
    "CYS": "Cysteine",
    "GLN": "Glutamine",
    "GLU": "Glutamic Acid",
    "GLY": "Glycine",
    "HIS": "Histidine",
    "HSD": "Histidine (Delta)",
    "HSE": "Histidine (Epsilon)",
    "HSP": "Histidine (Protonated)",
    "ILE": "Isoleucine",
    "LEU": "Leucine",
    "LYS": "Lysine",
    "MET": "Methionine",
    "PHE": "Phenylalanine",
    "PRO": "Proline",
    "SER": "Serine",
    "THR": "Threonine",
    "TRP": "Tryptophan",
    "TYR": "Tyrosine",
    "VAL": "Valine",
    "SEC": "Selenocysteine",
    "PYL": "Pyrrolysine",
}


class AminoAcid(Residue):
    """Class for Amino Acid Residue

    These objects should not be created by the user,
    but constructed automatically when parsing a
    coordinate file via amp.parsePDB or otherwise"""

    def __init__(
        self,
        name: str,
        index: int = None,
        number: int = None,
        chain: str = None,
        atoms=None,
    ):
        assert res_type(name) == "PRO"
        super(AminoAcid, self).__init__(name, index, number, chain, atoms)

        self._backbone = None
        self._sidechain = None
        self._sidechain_names = None

    @property
    def letter(self):
        """Amino acid alphabet"""
        return alphabet(self.name)

    @property
    def longname(self):
        """Amino acid chemical name"""
        return longname(self.name)

    @property
    def type(self):
        """Fixed typing"""
        return "PRO"

    @property
    def backbone(self):
        """Get backbone atoms

           Y
           |
        O==C   HN
           |   |
        R--CA--N--X
           |
           HA

        """

        if not self._backbone:
            self._backbone = [
                a for a in self.get_atom(BB_NAMES, verbosity=0) if a is not None
            ]

        return self._backbone

    def guess_hydrogen_names(self, verbosity=1):
        import mout
        import numpy as np

        for hydrogen in [a for a in self.atoms if a.symbol == "H"]:
            others = [
                a for a in self.atoms if a.symbol != "H" and a.index != hydrogen.index
            ]

            if len(others) == 0:
                mout.var("this_residue", self.name_number_str)
                self.plot3d()
                raise Exception(
                    "No other atoms in residue when guessing hydrogen names"
                )
            other_index = np.argmin([np.linalg.norm(a - hydrogen) for a in others])
            other = others[other_index]
            newname = f"H{other.name}"
            if verbosity:
                mout.warning(f"Renaming: {hydrogen} {hydrogen.number} --> {newname}")
            hydrogen.name = newname

    def remove_backbone(self, add_link=True, copy=False, verbosity=2):
        """Remove backbone atoms"""

        if copy:
            res = self.copy()
        else:
            res = self

        # res.summary()

        import mout

        if add_link:
            if verbosity:
                import mout

                mout.warningOut("Adding HLNK linker")
            import numpy as np

            vec_CB_CA = res.CA - res.CB
            unit_CB_CA = vec_CB_CA / np.linalg.norm(vec_CB_CA)
            target_length = 1.07  # from ase covalent radii
            res.CA.position = res.CB.np_pos + target_length * unit_CB_CA
            res.CA.set_name("HLNK", verbosity=verbosity - 1)

        for atom in res.backbone:
            if atom is None:
                continue
            res.delete_atom(atom.name, verbosity=verbosity - 1)

        return res

    def remove_sidechain(self, verbosity=2):
        """Remove sidechain atoms"""

        for atom in self.sidechain:
            if atom is None:
                continue
            self.delete_atom(atom.name, verbosity=verbosity - 1)

    @property
    def sidechain_names(self):
        if not self._sidechain_names:
            self._sidechain_names = [n for n in self.atom_names() if n not in BB_NAMES]

        return self._sidechain_names

    @property
    def sidechain(self):
        """Get sidechain atoms"""

        if not self._sidechain:
            self._sidechain = [
                a
                for a in self.get_atom(self.sidechain_names, verbosity=0)
                if a is not None
            ]

        return self._sidechain

    @property
    def CA(self):
        """Get Carbon-Alpha Atom"""
        return self.get_atom("CA", verbosity=0)

    @property
    def CB(self):
        """Get Carbon-Alpha Atom"""
        return self.get_atom("CB", verbosity=0)

    @property
    def HA(self):
        """Get Carbon-Alpha Hydrogen Atom"""
        return self.get_atom("HA", verbosity=0)

    @property
    def interaction_sites(self):

        import numpy as np
        from collections import namedtuple

        from .sites import Site

        sites = []

        for interaction in BB_INTERACTION_SITES + INTERACTION_SITES[self.name]:

            atoms = interaction["atoms"]

            if len(atoms) == 1:
                atoms = [self.get_atom(atoms[0])]
                position = atoms[0].np_pos
            else:
                atoms = self.get_atom(atoms)
                position = np.mean([a.np_pos for a in atoms], axis=0).round(3)

            # check if an equivalent site exists

            matches = [s for s in sites if np.linalg.norm(s.position - position) < 0.1]
            if len(matches) == 1:
                matches[0].types.append(interaction["type"])

            elif len(matches) == 0:
                sidechain = atoms[0].name in self.sidechain_names
                site = Site(
                    [interaction["type"]],
                    atoms,
                    position,
                    sidechain,
                    self.name,
                    self.number,
                    self.chain,
                )
                sites.append(site)

            else:
                raise Exception(
                    "Should never have multiple sites with the same position"
                )

        return sites

    def mutate(self, newname):
        """Mutate this amino acid to another"""

        import os
        import mout
        import mcol
        import glob

        mout.headerOut(
            f"Mutating {mcol.arg}{self.longname}{mcol.clear + mcol.bold} --> {mcol.arg}{longname(newname)}{mcol.clear + mcol.bold} ({mcol.result}{self.letter}{self.number}{alphabet(newname)}{mcol.clear + mcol.bold})"
        )

        # look for reference files
        amp_path = os.path.dirname(__file__)
        files = glob.glob(f"{amp_path}/ref/???.pdb")
        keys = [os.path.basename(f).replace(".pdb", "") for f in files]

        if newname == "ALA":
            from .mutate.protein import to_alanine

            to_alanine(self)

        elif newname in keys:

            from .io import parse
            from .mutate.protein import replace_sidechain

            reference = parse(f"{amp_path}/ref/{newname}.pdb").residues[0]
            replace_sidechain(self, reference)

        else:
            import mout

            mout.errorOut("Unsupported mutation!")

    @property
    def features(self):
        import mout
        import numpy as np
        from .rdkit import Feature

        features = []

        for f_dict in BB_FEATURES + FEATURES[self.name]:

            atoms = f_dict["atoms"]

            if len(atoms) == 1:
                atoms = [self.get_atom(atoms[0])]
                if atoms[0] is None:
                    # self.plot3d()
                    # mout.error(f'No atom {f_dict["atoms"][0]}', code='AminoAcid.features.0', fatal=True)
                    logger.warning(
                        f'No atom {f_dict["atoms"][0]}, is {self} {self.number} partially modelled?'
                    )
                    continue

                position = atoms[0].np_pos

            else:
                atoms = [a for a in self.get_atom(atoms) if a]
                if not atoms:
                    continue
                position = np.mean([a.np_pos for a in atoms], axis=0).round(3)

            feature = Feature(
                family=f_dict["family"],
                atoms=atoms,
                position=position,
                # sidechain=None,
                res_name=self.name,
                res_number=self.number,
                res_chain=self.chain,
            )

            # print(f_dict)
            features.append(feature)

        return features


"""

	Collate interactions per site

	Which interactions are shared among all amino acid backbones?

	Auto-checking rules:

		+ Hydrogen donor infers water donor
		+ Hydrogen acceptor infers water acceptor
		+ Most carbons should be able to form hydrophobic interactions
		+ Can hydrogen donors also be acceptors?


"""

RES_NAMES = [
    "ALA",
    "ARG",
    "ASN",
    "ASP",
    "CYS",
    "GLN",
    "GLU",
    "GLY",
    "HIS",
    "HSD",
    "HSE",
    "HSP",
    "ILE",
    "LEU",
    "LYS",
    "MET",
    "PHE",
    "PRO",
    "SER",
    "THR",
    "TRP",
    "TYR",
    "VAL",
]

BB_NAMES = ["N", "HN", "CA", "HA", "C", "O"]

BB_INTERACTION_SITES = [
    {"type": "water_donor", "atoms": ["N"], "source": "collated from ZV PLIP"},
    {"type": "hydrogen_donor", "atoms": ["N"], "source": "collated from ZV PLIP"},
    {"type": "water_acceptor", "atoms": ["O"], "source": "collated from ZV PLIP"},
    {"type": "hydrogen_acceptor", "atoms": ["O"], "source": "collated from ZV PLIP"},
]

BB_FEATURES = [
    dict(family="Donor", atoms=["N"], source="collated from ZV PLIP"),
    dict(family="Acceptor", atoms=["O"], source="collated from ZV PLIP"),
    dict(family="Donor", atoms=["O"], source="unsure"),
    dict(family="Acceptor", atoms=["N"], source="unsure"),
]

FEATURES = {
    "ALA": [
        dict(family="Hydrophobe", atoms=["CB"], source="ZV PLIP"),
    ],
    "ASN": [
        dict(family="Acceptor", atoms=["ND2"], source="MPro PLIP"),
        dict(family="Acceptor", atoms=["OD1"], source="unsure"),
        dict(family="Donor", atoms=["OD1"], source="MPro PLIP"),
        dict(family="Hydrophobe", atoms=["CB"], source="MPro PLIP"),
    ],
    "ASP": [
        dict(family="Acceptor", atoms=["OD1"], source="unsure"),
        dict(family="Donor", atoms=["OD1"], source="unsure"),
        dict(family="Acceptor", atoms=["OD2"], source="unsure"),
        dict(family="NegIonizable", atoms=["OD1", "OD2"], source="unsure"),
    ],
    "ASP": [
        dict(family="Acceptor", atoms=["OD1"], source="ZV PLIP"),
        dict(family="Donor", atoms=["OD1"], source="ZV PLIP"),
        dict(family="Acceptor", atoms=["OD2"], source="ZV PLIP"),
        dict(family="NegIonizable", atoms=["OD1", "OD2"], source="ZV PLIP"),
        dict(family="Hydrophobe", atoms=["CB"], source="ZV PLIP"),
    ],
    "ARG": [
        dict(family="Donor", atoms=["NH1"], source="unsure"),
        dict(family="Donor", atoms=["NH2"], source="unsure"),
        dict(family="Donor", atoms=["NE"], source="unsure"),
        dict(family="PosIonizable", atoms=["NH1"], source="unsure"),
        dict(family="PosIonizable", atoms=["NH2"], source="unsure"),
    ],
    "CYS": [
        dict(family="Hydrophobe", atoms=["CB"], source="unsure"),
        dict(family="PosIonizable", atoms=["SG"], source="unsure"),
        dict(family="Donor", atoms=["SG"], source="unsure"),
        dict(family="Sulfur", atoms=["SG"], source="unsure"),
    ],
    "GLN": [
        dict(family="Hydrophobe", atoms=["CG"], source="ZV PLIP"),
        dict(family="Acceptor", atoms=["NE2"], source="inferred (ASN)"),
        dict(family="Acceptor", atoms=["OE1"], source="inferred (ASN)"),
        dict(family="Donor", atoms=["OE1"], source="inferred (ASN)"),
        dict(family="Hydrophobe", atoms=["CB"], source="inferred (ASN)"),
    ],
    "GLU": [
        dict(family="Acceptor", atoms=["OE1"], source="ZV PLIP"),
        dict(family="Donor", atoms=["OE1"], source="ZV PLIP"),
        dict(family="Acceptor", atoms=["OE2"], source="ZV PLIP"),
        dict(family="NegIonizable", atoms=["OE1", "OE2"], source="ZV PLIP"),
        dict(family="Hydrophobe", atoms=["CB"], source="ZV PLIP"),
        dict(family="Hydrophobe", atoms=["CG"], source="ZV PLIP"),
    ],
    "GLY": [],
    "HIS": [
        dict(family="Hydrophobe", atoms=["CB"], source="ZV PLIP"),
        dict(
            family="Aromatic",
            atoms=["CG", "ND1", "CE1", "NE2", "CD2"],
            source="ZV PLIP",
        ),
        dict(family="Acceptor", atoms=["ND1"], source="unsure"),
        dict(family="Acceptor", atoms=["NE2"], source="unsure"),
        dict(family="Donor", atoms=["ND1"], source="unsure"),
        dict(family="Donor", atoms=["NE2"], source="unsure"),
    ],
    "HSE": [
        dict(family="Hydrophobe", atoms=["CB"], source="ZV PLIP"),
        dict(
            family="Aromatic",
            atoms=["CG", "ND1", "CE1", "NE2", "CD2"],
            source="ZV PLIP",
        ),
        dict(family="Acceptor", atoms=["ND1"], source="unsure"),
        dict(family="Acceptor", atoms=["NE2"], source="unsure"),
        dict(family="Donor", atoms=["NE2"], source="unsure"),
    ],
    "HSD": [
        dict(family="Hydrophobe", atoms=["CB"], source="ZV PLIP"),
        dict(
            family="Aromatic",
            atoms=["CG", "ND1", "CE1", "NE2", "CD2"],
            source="ZV PLIP",
        ),
        dict(family="Acceptor", atoms=["ND1"], source="unsure"),
        dict(family="Acceptor", atoms=["NE2"], source="unsure"),
        dict(family="Donor", atoms=["ND1"], source="unsure"),
    ],
    "HSP": [
        dict(family="Hydrophobe", atoms=["CB"], source="ZV PLIP"),
        dict(
            family="Aromatic",
            atoms=["CG", "ND1", "CE1", "NE2", "CD2"],
            source="ZV PLIP",
        ),
        dict(family="Acceptor", atoms=["ND1"], source="unsure"),
        dict(family="Acceptor", atoms=["NE2"], source="unsure"),
        dict(family="Donor", atoms=["ND1"], source="unsure"),
        dict(family="Donor", atoms=["NE2"], source="unsure"),
    ],
    "ILE": [
        dict(family="Hydrophobe", atoms=["CB"], source="ZV PLIP"),
        dict(family="Hydrophobe", atoms=["CD1"], source="ZV PLIP"),
        dict(family="Hydrophobe", atoms=["CG1"], source="ZV PLIP"),
        dict(family="Hydrophobe", atoms=["CG2"], source="ZV PLIP"),
    ],
    "LEU": [
        dict(family="Hydrophobe", atoms=["CB"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CD1"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CD2"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CG"], source="unsure"),
    ],
    "LYS": [
        dict(family="Hydrophobe", atoms=["CB"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CG"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CD"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CE"], source="unsure"),
        dict(family="PosIonizable", atoms=["NZ"], source="unsure"),
        dict(family="Donor", atoms=["NZ"], source="unsure"),
    ],
    "LYS": [
        dict(family="Hydrophobe", atoms=["CB"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CG"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CD"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CE"], source="unsure"),
        dict(family="PosIonizable", atoms=["NZ"], source="unsure"),
        dict(family="Donor", atoms=["NZ"], source="unsure"),
    ],
    "MET": [
        dict(family="Hydrophobe", atoms=["CE"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CG"], source="unsure"),
        dict(family="PosIonizable", atoms=["SD"], source="unsure"),
    ],
    "PHE": [
        dict(family="Hydrophobe", atoms=["CB"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CD1"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CD2"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CE1"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CE2"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CZ"], source="unsure"),
        dict(
            family="LumpedHydrophobe",
            atoms=["CG", "CD1", "CD2", "CE1", "CE2", "CZ"],
            source="unsure",
        ),
        dict(
            family="Aromatic",
            atoms=["CG", "CD1", "CD2", "CE1", "CE2", "CZ"],
            source="unsure",
        ),
    ],
    "PRO": [
        dict(family="Hydrophobe", atoms=["CG"], source="ZV PLIP"),
        dict(family="Aromatic", atoms=["N", "CA", "CB", "CG", "CD"], source="unsure"),
    ],
    "SER": [
        dict(family="Donor", atoms=["OG"], source="ZV PLIP"),
        dict(family="Acceptor", atoms=["OG"], source="ZV PLIP"),
    ],
    "THR": [
        dict(family="Donor", atoms=["OG1"], source="ZV PLIP"),
        dict(family="Acceptor", atoms=["OG1"], source="ZV PLIP"),
        dict(family="Hydrophobe", atoms=["CG2"], source="ZV PLIP"),
    ],
    "TRP": [
        dict(family="Hydrophobe", atoms=["CZ2"], source="ZV PLIP"),
        dict(family="Hydrophobe", atoms=["CH2"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CZ3"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CE3"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CB"], source="ZV PLIP"),
        dict(
            family="Aromatic", atoms=["CG", "CD1", "NE1", "CE2", "CD2"], source="unsure"
        ),
        dict(
            family="Aromatic",
            atoms=["CE2", "CD2", "CZ2", "CZ3", "CE3", "CH2"],
            source="unsure",
        ),
        dict(
            family="LumpedHydrophobe",
            atoms=["CE2", "CD2", "CZ2", "CZ3", "CE3", "CH2"],
            source="unsure",
        ),
    ],
    "TYR": [
        dict(
            family="Aromatic",
            atoms=["CG", "CD1", "CE1", "CZ", "CE2", "CD2"],
            source="ZV PLIP",
        ),
        dict(
            family="LumpedHydrophobe",
            atoms=["CG", "CD1", "CE1", "CZ", "CE2", "CD2"],
            source="ZV PLIP",
        ),
        dict(family="Hydrophobe", atoms=["CD1"], source="ZV PLIP"),
        dict(family="Hydrophobe", atoms=["CD2"], source="ZV PLIP"),
        dict(family="Hydrophobe", atoms=["CE1"], source="ZV PLIP"),
        dict(family="Hydrophobe", atoms=["CE2"], source="ZV PLIP"),
        dict(family="Hydrophobe", atoms=["CG"], source="ZV PLIP"),
        dict(family="Hydrophobe", atoms=["CB"], source="ZV PLIP"),
        dict(family="Donor", atoms=["OH"], source="ZV PLIP"),
        dict(family="Acceptor", atoms=["OH"], source="ZV PLIP"),
    ],
    "VAL": [
        dict(family="Hydrophobe", atoms=["CB"], source="unsure"),
        dict(family="Hydrophobe", atoms=["CG1"], source="ZV PLIP (inferred)"),
        dict(family="Hydrophobe", atoms=["CG2"], source="ZV PLIP"),
    ],
}

INTERACTION_SITES = {
    "ALA": [
        {"type": "hydrophobic", "atoms": ["CB"], "source": "ZV PLIP"},
    ],
    "ASN": [
        # {"type": "hydrophobic", "atoms": ["CB"], "source": "MPro PLIP"},
        # {"type": "hydrophobic", "atoms": ["CG"], "source": "MPro PLIP"},
        # {"type": "water_acceptor", "atoms": ["ND2"], "source": "MPro PLIP"},
        # {"type": "hydrogen_donor", "atoms": ["ND2"], "source": "unsure"},
    ],
    "ASP": [
        {"type": "hydrogen_acceptor", "atoms": ["OD2"], "source": "ZV PLIP"},
        {"type": "water_acceptor", "atoms": ["OD2"], "source": "ZV PLIP (inferred)"},
        {"type": "hydrogen_acceptor", "atoms": ["OD1"], "source": "ZV PLIP"},
        {"type": "hydrogen_acceptor", "atoms": ["CG"], "source": "ZV PLIP"},
        {"type": "water_acceptor", "atoms": ["CG"], "source": "ZV PLIP (inferred)"},
        {"type": "salt_bridge", "atoms": ["OD1", "OD2"], "source": "ZV PLIP"},
        {"type": "water_acceptor", "atoms": ["OD1"], "source": "ZV PLIP"},
        {"type": "hydrophobic", "atoms": ["CB"], "source": "ZV PLIP"},
        {"type": "hydrogen_donor", "atoms": ["OD1"], "source": "ZV PLIP"},
        {"type": "water_donor", "atoms": ["OD1"], "source": "ZV PLIP (inferred)"},
    ],
    "ARG": [
        {"type": "hydrogen_donor", "atoms": ["NH1"], "source": "unsure"},
        {"type": "hydrogen_donor", "atoms": ["NH2"], "source": "unsure"},
        {"type": "water_donor", "atoms": ["NH1"], "source": "unsure"},
        {"type": "water_donor", "atoms": ["NH2"], "source": "unsure"},
        {"type": "salt_bridge", "atoms": ["NH1"], "source": "unsure"},
        {"type": "salt_bridge", "atoms": ["NH2"], "source": "unsure"},
        {"type": "pi_cation", "atoms": ["NH1"], "source": "unsure"},
        {"type": "pi_cation", "atoms": ["NH2"], "source": "unsure"},
    ],
    "CYS": [
        {"type": "pi_cation", "atoms": ["SG"], "source": "unsure"},
        {"type": "sulfur_sulfur", "atoms": ["SG"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CB"], "source": "unsure"},
    ],
    "GLN": [
        # {"type": "hydrophobic", "atoms": ["CB"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CG"], "source": "ZV PLIP"},
        # {"type": "hydrogen_donor", "atoms": ["NE2"], "source": "unsure"},
        # {"type": "hydrogen_acceptor", "atoms": ["OE1"], "source": "unsure"},
        # {"type": "halogen_acceptor", "atoms": ["O2"], "source": "MPro PLIP"},
    ],
    "GLU": [
        {"type": "hydrogen_donor", "atoms": ["OE1"], "source": "ZV PLIP"},
        {"type": "water_donor", "atoms": ["OE1"], "source": "ZV PLIP (inferred)"},
        # {"type": "hydrogen_donor", "atoms": ["CA"], "source": "MPro PLIP"},
        {"type": "hydrogen_acceptor", "atoms": ["OE1"], "source": "unsure"},
        {"type": "water_acceptor", "atoms": ["OE1"], "source": "unsure"},
        # {"type": "hydrogen_donor", "atoms": ["N"], "source": "MPro PLIP"},
        # {"type": "hydrophobic", "atoms": ["CB"], "source": "MPro PLIP"},
    ],
    "GLY": [
        # {"type": "water_donor", "atoms": ["O"]},
    ],
    "HIS": [
        {"type": "hydrophobic", "atoms": ["CB"], "source": "ZV PLIP"},
        {
            "type": "pi_stacking",
            "atoms": ["CG", "ND1", "CE1", "NE2", "CD2"],
            "source": "ZV PLIP",
        },
        {
            "type": "pi_cation",
            "atoms": ["CG", "ND1", "CE1", "NE2", "CD2"],
            "source": "ZV PLIP (inferred)",
        },
    ],
    "HSE": [
        {"type": "hydrophobic", "atoms": ["CB"], "source": "ZV PLIP (inferred)"},
        {
            "type": "pi_stacking",
            "atoms": ["CG", "ND1", "CE1", "NE2", "CD2"],
            "source": "ZV PLIP (inferred)",
        },
        {
            "type": "pi_cation",
            "atoms": ["CG", "ND1", "CE1", "NE2", "CD2"],
            "source": "ZV PLIP (inferred)",
        },
    ],
    "HSD": [
        {"type": "hydrophobic", "atoms": ["CB"], "source": "ZV PLIP (inferred)"},
        {
            "type": "pi_stacking",
            "atoms": ["CG", "ND1", "CE1", "NE2", "CD2"],
            "source": "ZV PLIP (inferred)",
        },
        {
            "type": "pi_cation",
            "atoms": ["CG", "ND1", "CE1", "NE2", "CD2"],
            "source": "ZV PLIP (inferred)",
        },
    ],
    "HSP": [
        {"type": "hydrophobic", "atoms": ["CB"], "source": "ZV PLIP (inferred)"},
        {
            "type": "pi_stacking",
            "atoms": ["CG", "ND1", "CE1", "NE2", "CD2"],
            "source": "ZV PLIP (inferred)",
        },
        {
            "type": "pi_cation",
            "atoms": ["CG", "ND1", "CE1", "NE2", "CD2"],
            "source": "ZV PLIP (inferred)",
        },
    ],
    "ILE": [
        {"type": "hydrophobic", "atoms": ["CD1"], "source": "ZV PLIP"},
        {"type": "hydrophobic", "atoms": ["CB"], "source": "ZV PLIP"},
    ],
    "LEU": [
        {"type": "hydrophobic", "atoms": ["CB"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CG"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CD1"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CD2"], "source": "unsure"},
    ],
    "LYS": [
        {"type": "hydrogen_donor", "atoms": ["NZ"], "source": "unsure"},
        {"type": "water_donor", "atoms": ["NZ"], "source": "unsure"},
        {"type": "salt_bridge", "atoms": ["NZ"], "source": "unsure"},
        {"type": "pi_cation", "atoms": ["NZ"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CB"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CG"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CD"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CE"], "source": "unsure"},
    ],
    "MET": [
        {"type": "pi_cation", "atoms": ["SD"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CE"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CG"], "source": "unsure"},
        # {"type": "hydrophobic", "atoms": ["CB"], "source": "MPro PLIP"},
    ],
    "PHE": [
        {"type": "hydrophobic", "atoms": ["CB"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CG"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CD1"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CD2"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CE1"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CE2"], "source": "unsure"},
        {"type": "hydrophobic", "atoms": ["CZ"], "source": "unsure"},
        {
            "type": "pi_stacking",
            "atoms": ["CG", "CD1", "CD2", "CE1", "CE2", "CZ"],
            "source": "unsure",
        },
        {
            "type": "pi_cation",
            "atoms": ["CG", "CD1", "CD2", "CE1", "CE2", "CZ"],
            "source": "unsure",
        },
    ],
    "PRO": [
        {"type": "hydrophobic", "atoms": ["CG"], "source": "ZV PLIP"},
    ],
    "SER": [
        {"type": "hydrogen_acceptor", "atoms": ["OG"], "source": "ZV PLIP"},
        {"type": "water_acceptor", "atoms": ["OG"], "source": "ZV PLIP (inferred)"},
        {"type": "hydrogen_donor", "atoms": ["OG"], "source": "ZV PLIP"},
        {"type": "water_donor", "atoms": ["OG"], "source": "ZV PLIP (inferred)"},
        # {"type": "hydrogen_donor", "atoms": ["CB"], "source": "MPro PLIP"},
    ],
    "THR": [
        {"type": "hydrophobic", "atoms": ["CG2"], "source": "ZV PLIP"},
        {"type": "hydrogen_donor", "atoms": ["OG1"], "source": "ZV PLIP"},
        {"type": "water_donor", "atoms": ["OG1"], "source": "ZV PLIP (inferred)"},
        {"type": "hydrogen_acceptor", "atoms": ["OG1"], "source": "ZV PLIP"},
        {"type": "water_acceptor", "atoms": ["OG1"], "source": "ZV PLIP"},
        # {"type": "water_acceptor", "atoms": ["CB"]},
    ],
    "TRP": [
        {"type": "hydrophobic", "atoms": ["CZ2"], "source": "ZV PLIP"},
        {"type": "hydrophobic", "atoms": ["CB"], "source": "ZV PLIP"},
        {
            "type": "pi_stacking",
            "atoms": ["CG", "CD1", "CD2", "NE1", "CE2"],
            "source": "ZV PLIP",
        },
        {
            "type": "pi_cation",
            "atoms": ["CG", "CD1", "CD2", "NE1", "CE2"],
            "source": "ZV PLIP (inferred)",
        },
    ],
    "TYR": [
        {"type": "hydrophobic", "atoms": ["CB"], "source": "ZV PLIP"},
        {
            "type": "pi_stacking",
            "atoms": ["CG", "CD1", "CE1", "CZ", "CE2", "CD2"],
            "source": "ZV PLIP",
        },
        {"type": "hydrophobic", "atoms": ["CE1"], "source": "ZV PLIP"},
        {"type": "hydrophobic", "atoms": ["CD1"], "source": "ZV PLIP"},
        {"type": "hydrophobic", "atoms": ["CE2"], "source": "ZV PLIP"},
        {"type": "hydrophobic", "atoms": ["CG"], "source": "ZV PLIP"},
        {
            "type": "pi_cation",
            "atoms": ["CG", "CD1", "CE1", "CZ", "CE2", "CD2"],
            "source": "ZV PLIP",
        },
    ],
    "VAL": [
        {"type": "hydrophobic", "atoms": ["CG2"], "source": "ZV PLIP"},
    ],
}



def compare_mols(m1, m2, names=["m1", "m2"], ignore_hydrogen=False):

    from rdkit.Chem.rdMolDescriptors import CalcMolFormula

    f1 = CalcMolFormula(m1)
    f2 = CalcMolFormula(m2)

    from molparse.atomtypes import formula_to_atomtype_dict, subtract_atomtype_dict

    d1 = formula_to_atomtype_dict(f1)
    d2 = formula_to_atomtype_dict(f2)

    diff1 = subtract_atomtype_dict(d1, d2, ignore_hydrogen=ignore_hydrogen)
    diff2 = subtract_atomtype_dict(d2, d1, ignore_hydrogen=ignore_hydrogen)

    return diff1, diff2



def peakFinder(
    xdata,
    ydata,
    min_width,
    min_height,
    baseline=None,
    threshold=1.0e-2,
    search_coeff=0.1,
):
    import statistics
    import mout
    from .modify import differentiate

    many = any(isinstance(el, list) for el in ydata)

    xdata, dydx = differentiate(xdata, ydata)

    if many:

        peaklist = []

        for data in ydata:
            peaklist.append(
                peakFinder(
                    xdata,
                    data,
                    min_width,
                    min_height,
                    baseline=baseline,
                    threshold=threshold,
                    search_coeff=search_coeff,
                )
            )

        all_peaks = peaklist

    else:

        if baseline is None:
            y_avg = statistics.mean(ydata)
        else:
            y_avg = baseline

        # print(y_avg)

        found = False

        all_peaks = []

        for i, y in enumerate(ydata):

            this_height = y - y_avg

            if this_height > min_height and abs(dydx[i]) < threshold:

                # print(xdata[i],y,this_height,abs(dydx[i]))

                skip = False
                for peak in all_peaks:
                    if peak.in_range(xdata[i]):
                        skip = True
                        break

                if skip:
                    continue

                # backwards search for the start of the peak
                backsearch_i = i
                backsearch_height = this_height

                while backsearch_height > min_height * search_coeff:

                    backsearch_i -= 1
                    if backsearch_i <= 0:
                        break
                    # print(backsearch_i)
                    backsearch_height = ydata[backsearch_i] - y_avg

                mout.varOut("Peak start estimated at", xdata[backsearch_i])

                # forwards search for the start of the peak
                foresearch_i = i
                foresearch_height = this_height

                while foresearch_height > min_height * search_coeff:

                    foresearch_i += 1
                    if foresearch_i >= len(ydata) - 1:
                        break
                    foresearch_height = ydata[foresearch_i] - y_avg

                mout.varOut("Peak end estimated at", xdata[foresearch_i])

                all_peaks.append(Peak(xdata[backsearch_i], xdata[foresearch_i]))

    return all_peaks


class Peak:
    def __init__(self, start_x, end_x):

        self.start_x = start_x
        self.end_x = end_x

    def __str__(self):
        return "Peak[" + str(self.start_x) + ":" + str(self.end_x) + "]"

    def __repr__(self):
        return "Peak[" + str(self.start_x) + ":" + str(self.end_x) + "]"

    def in_range(self, x):
        if x >= self.start_x and x <= self.end_x:
            return True
        else:
            return False

    def gauss_fit(self, xdata, ydata, return_data=False, baseline=0.0, filename=None):
        return gaussian_fit(
            xdata,
            ydata,
            [self.start_x, self.end_x],
            return_data=return_data,
            baseline=baseline,
            filename=filename,
        )


def gaussian_fit(
    xdata,
    ydata,
    window,
    return_data=False,
    index_window=False,
    baseline=0.0,
    filename=None,
):
    import mout
    from scipy.optimize import curve_fit
    from scipy import asarray as ar

    many = any(isinstance(el, list) for el in ydata)

    if many:
        mout.errorOut("gaussian_fit. Unsupported.", fatal=True)

    if index_window:
        x = ar(xdata[window[0] : window[1]])
        y = ar(ydata[window[0] : window[1]])
    else:
        index1 = closest_index(window[0], xdata)
        index2 = closest_index(window[1], xdata)
        x = ar(xdata[index1:index2])
        y = ar(ydata[index1:index2])

    n = len(x)  # the number of data
    mean = sum(x * y) / n
    sigma = sum(y * (x - mean) ** 2) / n

    mean = x[len(x) // 2]
    sigma = (x[-1] - x[0]) / 4

    import mplot

    # print(1.0,mean,sigma)

    global ___BASELINE
    ___BASELINE = baseline

    try:
        popt, pcov = curve_fit(gaus, x, y, p0=[1, mean, sigma])
    except:
        mplot.graph2D(x, y, filename=filename, show=False)
        return False

    a, mean, sigma = popt

    # print(a,mean,sigma)

    if return_data:
        gaus_y = []
        for this_x in xdata:
            gaus_y.append(gaus(this_x, *popt))
        if filename is not None:
            mplot.graph2D(x, [y, gaus_y[index1:index2]], filename=filename, show=False)
        return a, mean, sigma, xdata, gaus_y
    else:
        return a, mean, sigma


def gaus(x, a, x0, sigma):
    from scipy import asarray as exp

    return a * exp(-((x - x0) ** 2) / (2 * sigma**2)) + ___BASELINE


# def closest_index(value,xdata):
# 	"""Return the index of the nearest data point in the array"""
# 	if isinstance(value,list):
# 		result = []
# 		for v in value:
# 			result.append(closest_index(v, xdata))
# 		return result
# 	else:
# 		for i,x in enumerate(xdata):
# 			if x > value:
# 				if abs(xdata[i-1]-value) < abs(xdata[i]-value):
# 					return i
# 				else:
# 					return i-1


def closest_index(value, xdata, comparison=None, numpy=False):
    """Return the index of the nearest data point in the array"""
    if numpy:
        if comparison:
            import mout

            mout.warningOut(
                "numpy argument overrides comparison",
                code="amp.signal.peaks.closest_index[1]",
            )
        import numpy as np

        array = np.asarray(xdata)
        return (np.abs(array - value)).argmin()
    elif isinstance(value, list):
        result = []
        for v in value:
            result.append(closest_index(v, xdata, comparison=comparison))
        return result
    else:
        for i, x in enumerate(xdata):
            if (i > 0 and x < xdata[i - 1]) or (i == 0 and x > xdata[i + 1]):
                import mout

                mout.errorOut("Array does not appear to be sorted!")
                return -1
            elif x > value:
                if comparison is not None:
                    if comparison == "<":
                        return i - i
                    else:
                        return i
                else:
                    if abs(xdata[i - 1] - value) < abs(xdata[i] - value):
                        return i
                    else:
                        return i - 1
        return i


def closest_value(
    xvalue, xdata, ydata=None, comparison=None, numpy=False, return_x=False
):
    """Return the yvalue of the data point nearest the given xvalue"""
    if ydata is None:
        ydata = xdata
    if isinstance(xvalue, list):
        x_result = []
        y_result = []
        for x in xvalue:
            xres, yres = closest_value(
                x, xdata, ydata, comparison=comparison, numpy=numpy, return_x=return_x
            )
            x_result = []
            y_result = []
        if return_x:
            return x_result, y_result
        else:
            return y_result
    else:
        index = closest_index(xvalue, xdata, comparison=comparison, numpy=numpy)
        if return_x:
            return xdata[index], ydata[index]
        else:
            return ydata[index]



import py3Dmol


def render(
    sys,
    protein="cartoon",
    ligand="stick",
    protein_color="spectrum",
    width="640px",
    height="480px",
):

    from tempfile import NamedTemporaryFile

    # protein
    view = py3Dmol.view(data=sys.protein_system.pdb_block, width=width, height=height)
    view.setStyle({protein: {"color": protein_color}})

    # ligands
    if ligand:
        for res in sys.ligand_residues:
            view.addModel(res.pdb_block, "pdb")
            view.setStyle({"model": -1}, {ligand: {}})

    return view



from .group import AtomGroup


# from enum import Enum
# class UnknownSiteHandlingMethod(Enum):
#   IGNORE = 0 # atoms with no alternative sites are ignored when separating
#   UNION = 1  # all atoms with no alternative sites are added to each separated residue


class Residue(AtomGroup):
    """Class for chemical Residue

    These objects should not be created by the user,
    but constructed automatically when parsing a
    coordinate file via amp.parsePDB or otherwise"""

    def __init__(
        self,
        name: str,
        index: int = None,
        number: int = None,
        chain: str = None,
        atoms=None,
    ):

        super(Residue, self).__init__(name)

        self.chain = chain
        self.index = index
        self._number = number

        from .list import NamedList

        self._atoms = NamedList()

        if atoms:

            from .atom import Atom

            for atom in atoms:
                self.add_atom(
                    Atom(
                        name=atom.name,
                        index=atom.index,
                        pdb_index=atom.number,
                        position=atom.position,
                        residue=self.name,
                        res_number=self.number,
                        chain=self.chain,
                    )
                )

        self._parent = None

    @property
    def alternative_sites(self):
        if self.contains_alternative_sites:
            return list(
                set(
                    [
                        a.alternative_site
                        for a in self.atoms
                        if a.alternative_site is not None
                    ]
                )
            )

        return []

    def split_by_site(self, unknown_site_handling="ignore"):
        """return a list of Residue objects, one for each site"""
        if self.contains_alternative_sites:
            ignore_count = 0
            separated_residues = []
            import mout

            # mout.debug(f'{self.alternative_sites=}')
            for site in self.alternative_sites:

                res = Residue(
                    self.name, index=self.index, number=self.number, chain=self.chain
                )
                for atom in self.atoms:
                    if atom.alternative_site == site:
                        res.add_atom(atom)
                    elif atom.alternative_site is None:
                        if unknown_site_handling == "union":
                            res.add_atom(atom)
                        else:
                            ignore_count += 1

                separated_residues.append(res)

            if ignore_count:
                import mout

                mout.warning(
                    f"Ignored {ignore_count} atoms with no alternative site [{self.name_number_str}]"
                )

            return separated_residues
        else:
            return [self]

    @property
    def name_number_str(self):
        return f"{self.name} {self.number}"

    @property
    def name_number_chain_str(self):
        return f"{self.name} {self.number} {self.chain}"

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, obj):
        # import weakref
        # self._parent = weakref.ref(obj)
        self._parent = obj

    def rename(self, new: str, verbosity: int = 1):
        """Rename the residue"""
        if verbosity > 0:
            import mcol
            import mout

            mout.out(
                "Renaming residue "
                + mcol.arg
                + self.name
                + str([self.number])
                + mcol.clear
                + " of chain "
                + mcol.varName
                + self.chain
                + mcol.clear
                + " to "
                + mcol.arg
                + new
            )
        self._name = new
        self.fix_names()

    @property
    def resid(self):
        """resid and number are the residue index from the original file"""
        return self._number

    @property
    def number(self):
        """resid and number are the residue index from the original file"""
        return self._number

    @number.setter
    def number(self, a):
        """resid and number are the residue index from the original file"""
        self._number = a
        for atom in self.atoms:
            atom._res_number = a

    @property
    def ase_atoms(self):
        from ase import Atoms

        atoms = self.atoms
        symbols = [a.symbol for a in atoms]
        positions = [a.position for a in atoms]
        return Atoms(symbols=symbols, cell=None, pbc=None, positions=positions)

    def view(self):
        """View the system with ASE"""
        from .gui import view

        view(self.ase_atoms)

    def fix_names(self):
        """Ensure child Atoms have correct parent name"""
        for atom in self.atoms:
            atom.residue = self.name

    def fix_indices(self):
        """Ensure child Atoms have correct res_index"""
        for atom in self.atoms:
            atom.res_index = self.index

    def atom_names(self, wRes: bool = False, noPrime: bool = False):
        """Returns names of all child Atoms (list)"""
        names_list = []
        for atom in self._atoms:
            names_list.append(atom.get_name(wRes=wRes, noPrime=noPrime))
        return names_list

    @property
    def atomic_numbers(self):
        """Returns atomic numbers of all child Atoms (list)"""
        number_list = []
        for atom in self._atoms:
            number_list.append(atom.atomic_number)
        return number_list

    @property
    def FF_atomtypes(self):
        """Returns atomtypes of all child Atoms (list)"""
        atomtype_list = []
        for atom in self._atoms:
            atomtype_list.append(atom.FF_atomtype)
        return atomtype_list

    @property
    def type(self):
        """Guess type from residue name"""
        return res_type(self.name)

    @property
    def letter(self):
        return "?"

    def add_atom(self, atom):
        """add an Atom"""
        self.addAtom(atom)

    def addAtom(self, atom, copy=True):
        """add an Atom"""
        from .atom import Atom

        assert isinstance(atom, Atom)

        if copy:
            atom_new = atom.copy()
            # import mout
            # mout.debug(f'copied atom: {atom} {id(atom)} --> {id(atom_new)}')
            atom = atom_new

        atom.chain = self.chain
        atom.residue = self.name
        atom.res_index = self.index
        atom._res_number = self.number

        atom.parent = self

        if self.atoms:
            # remove any termini
            self.atoms[-1].terminal = None
            self.atoms[-1].ter_line = None

        self._atoms.append(atom)

    def translate(self, vector):
        """Adjust the position of all atoms in the residue"""
        for atom in self._atoms:
            atom.position = atom.np_pos + vector

    def print(self):
        """Verbose output of the Residue's properties"""
        import mout

        mout.varOut("Residue Name", self.name)
        mout.varOut("Residue Chain", self.chain)
        mout.varOut("Residue Number", self.number)
        mout.varOut("Number of Atoms", self.num_atoms)

    @property
    def atoms(self):
        """Returns all child Atoms (list)"""
        return self._atoms

    def get_atom(self, name, verbosity=1):
        """Get a child Atom by name"""
        if isinstance(name, list):
            return [self.get_atom(n, verbosity=verbosity - 1) for n in name]

        for atom in self._atoms:
            if atom.name == name:
                return atom

        import mout
        import mcol

        if verbosity > 0:
            mout.errorOut(
                "No atom "
                + mcol.arg
                + name
                + mcol.error
                + " in residue "
                + mcol.arg
                + self.name
                + str([self.number]),
                fatal=False,
                code="Residue.1",
            )
        return None

    def set_positions(self, positions):
        """Set positions for all child atoms"""
        for atom, pos in zip(self.atoms, positions):
            atom.position = pos

    def align_to(self, target, names):
        """Minimise rotation w.r.t. a target"""

        import numpy as np
        from ase.build.rotate import rotation_matrix_from_points

        assert isinstance(target, Residue)

        # get subset of the residue to align
        self_copy = self.copy()
        self_copy.delete_atom([n for n in self_copy.atom_names() if n not in names])

        # get subset of the target
        target_copy = target.copy()
        target_copy.delete_atom([n for n in target_copy.atom_names() if n not in names])

        # position lists
        p = self_copy.ase_atoms.get_positions()
        p0 = target_copy.ase_atoms.get_positions()

        # centeroids
        c = np.mean(p, axis=0)
        c0 = np.mean(p0, axis=0)

        # subtract centroids
        p -= c
        p0 -= c0

        # Compute rotation matrix
        R = rotation_matrix_from_points(p.T, p0.T)

        # compute displacement
        p_new = np.dot(self_copy.ase_atoms.get_positions(), R.T) + c0
        d = p0[0] + c0 - p_new[0]

        # set the new positions
        self.set_positions(np.dot(self.ase_atoms.get_positions(), R.T) + c0 + d)

    def delete_atom(self, name: str, verbosity: int = 1):
        """Delete a child Atom by name"""
        import mcol
        import mout

        if isinstance(name, list):
            for n in name:
                self.delete_atom(n, verbosity=verbosity - 1)
            return
        for index, atom in enumerate(self._atoms):
            if atom.name == name:
                del self._atoms[index]
                if verbosity > 0:
                    mout.warningOut(
                        "Deleted "
                        + mcol.result
                        + "1"
                        + mcol.warning
                        + " atom of name "
                        + mcol.arg
                        + name
                        + mcol.warning
                        + " in residue "
                        + mcol.arg
                        + self.name
                    )
                return
        if verbosity > 1:
            mout.warningOut(
                "Deleted "
                + mcol.result
                + "0"
                + mcol.warning
                + " atom of name "
                + mcol.arg
                + name
                + mcol.warning
                + " in residue "
                + mcol.arg
                + self.name
            )

    def copy(self, fast: bool = False):
        """Return a deepcopy of the Residue"""
        if fast:
            new_residue = Residue(self.name, self.number, self.chain)
            for atom in self.atoms:
                new_residue.addAtom(atom.copy(fast=fast))
            return new_residue
        else:
            import copy

            return copy.deepcopy(self)

    def get_distance(self, i: str, j: str):
        """Get distance between two named atoms"""
        if type(i) is str:
            i = self.get_atom(name=i)
        elif type(i) is int:
            i = self._atoms[i]
        if type(j) is str:
            j = self.get_atom(name=j)
        elif type(j) is int:
            j = self._atoms[j]

        return displacement(i.position, j.position)

    def get_angle(self, i: str, j: str, k: str):
        """Get angle between three named atoms"""
        if type(i) is str:
            i = self.get_atom(name=i)
        elif type(i) is int:
            i = self._atoms[i]
        if type(j) is str:
            j = self.get_atom(name=j)
        elif type(j) is int:
            j = self._atoms[j]
        if type(k) is str:
            k = self.get_atom(name=k)
        elif type(k) is int:
            k = self._atoms[k]

        return angle(i.position, j.position, k.position)

    def summary(self):
        """Summarised output of the Residue"""
        import mout
        import mcol

        mout.headerOut(
            f"\nResidue {self.name}, index={self.index}, number={self.number}, chain={self.chain}, #atoms={self.num_atoms}"
        )

        mout.out(
            f"{mcol.underline}{'Sy':2} {'NAME':4} {'INDEX':>6} {'NUMBER':>6} {'X':>7} {'Y':>7} {'Z':>7} {'Alt.':>4}{mcol.clear} "
        )
        for atom in self.atoms:
            mout.out(
                f'{atom.symbol:2} {atom.name:4} {atom.index if atom.index is not None else " ":>6} {atom.number if atom.number is not None else " ":>6} {atom.x:>7.2f} {atom.y:>7.2f} {atom.z:>7.2f} {atom.alternative_site or " ":>4}'
            )

    def is_same_as(self, residue):
        assert isinstance(residue, Residue)

        if self.name != residue.name:
            return False
        if self.number != residue.number:
            return False
        if self.chain != residue.chain:
            return False
        return True

    def set_chain_number(self, index):
        for atom in self.atoms:
            atom.chain_number = index

    def set_chain_char(self, char):
        for atom in self.atoms:
            atom.chain = char

    @property
    def children(self):
        return self.atoms

    def prune_alternative_sites(self, site="A", verbosity=1):
        """Remove atoms with alternative sites"""
        import mout

        count = 0
        delete_list = []
        for index, atom in enumerate(self.atoms):
            if not atom.alternative_site:
                continue
            elif atom.alternative_site == site:
                atom._alternative_site = None
            elif atom.alternative_site != site:
                delete_list.append(index)
                count += 1

        for index in reversed(delete_list):
            del self._atoms[index]

        if verbosity > 0 and count > 0:
            mout.warningOut(f"Deleted {count} alternative site atoms")

        return count


RES_TYPES = {
    "DNA": ("DA", "DT", "DC", "DG", "ADE9", "THMN", "GUA9", "CTSN", "DOG"),
    "SOL": ("SOL", "WAT", "TIP", "T3P", "HOH", "PEG", "SO4", "DMS", "H2S"),
    "ION": (
        "ION",
        "MG",
        "CL",
        "NA",
        "SOD",
        "POT",
        "CAL",
        "LIT",
        "Na+",
        "Cl-",
        "CA",
        "ZN",
    ),
    "LIP": ("DPPC", "POPC", "DAG", "TAG"),
    "LIG": ("ATP", "GTP", "LIG", "UNL", "CTP", "TTP", "OGTP", "TRS", "APR"),
    "N/A": ("QM", "MM"),
    "PRO": (
        "ALA",
        "ARG",
        "ASN",
        "ASP",
        "CYS",
        "GLN",
        "GLU",
        "GLY",
        "HSD",
        "HSE",
        "HIS",
        "ILE",
        "LEU",
        "LYS",
        "MET",
        "PHE",
        "PRO",
        "SER",
        "THR",
        "TRP",
        "TYR",
        "VAL",
        "HID",
        "HIE",
        "HIP",
        "HSP",
    ),
}

EXTRA_TYPES = {}

WARNED_TYPES = []

WARN_UNKNOWN_RES = True


def add_res_type(resname, type):
    global EXTRA_TYPES
    EXTRA_TYPES[resname] = type


def silence_unknown_res_warning():
    global WARN_UNKNOWN_RES
    WARN_UNKNOWN_RES = False


def res_type(resname):
    """Guess type from residue name"""

    global WARNED_TYPES

    if resname in EXTRA_TYPES:
        return EXTRA_TYPES[resname]

    for k, v in RES_TYPES.items():
        if resname in v:
            return k

    # Amber style amino acid termini
    if (
        resname[0] in ("N", "C")
        and len(resname) == 4
        and resname[1:] in RES_TYPES["PRO"]
    ):
        return "PRO"

    # Chromophore
    if resname in ["SYG", "CRO", "NTYG", "DGLU", "AS2T", "SYTY", "ASYG"]:
        return "PRO"

    import mcol
    import mout

    if WARN_UNKNOWN_RES and resname not in WARNED_TYPES:
        mout.warningOut(
            f"Unknown residue type for {mcol.arg}{resname}{mcol.warning}. Defaulted to {mcol.arg}LIG"
        )
        WARNED_TYPES.append(resname)
    return "LIG"



"""

To-Do's
  * If show = True close all other figures?
  * graphVelocity
  * graphTemperature
  * graphGyration

"""


def graphEnergy(
    trajectory,
    perAtom=True,
    filename=None,
    show=True,
    verbosity=2,
    kJpermol=False,
    xlab=None,
    timestep=None,
    xmin=None,
    xmax=None,
    ymin=None,
    ymax=None,
):
    import mcol
    import mout
    import mplot
    from ase import units

    if verbosity > 0:
        mout.out(
            "graphing " + mcol.varName + "Energy" + mcol.clear + " ... ",
            printScript=True,
            end="",
        )  # user output

    xdata = []
    ekins = []
    epots = []
    etots = []

    if not kJpermol:
        if perAtom:
            ylab = "Energy [eV/atom]"
        else:
            ylab = "Energy [eV]"
    else:
        ylab = "Energy [kJ/mol]"

    for n, atoms in enumerate(trajectory):
        if timestep is None:
            if xlab is None:
                xlab = "MD Steps"
            xdata.append(n)
        else:
            xlab = "Time [fs]"
            xdata.append(n * timestep / units.fs)
        epot = atoms.get_potential_energy()
        ekin = atoms.get_kinetic_energy()

        if not kJpermol:
            if perAtom:
                epot /= len(atoms)
                ekin /= len(atoms)
                ylab = "Energy [eV/atom]"
            else:
                ylab = "Energy [eV]"
        else:
            epot *= units.eV / units.kJ * units.mol
            ekin *= units.eV / units.kJ * units.mol
            ylab = "Energy [kJ/mol]"

        epots.append(epot)
        ekins.append(ekin)
        etots.append(epot + ekin)

    mplot.graph2D(
        xdata,
        [epots, ekins, etots],
        ytitles=["Potential", "Kinetic", "Total"],
        show=show,
        xlab=xlab,
        ylab=ylab,
        filename=filename,
        verbosity=verbosity - 1,
        xmin=xmin,
        xmax=xmax,
        ymin=ymin,
        ymax=ymax,
    )

    if verbosity > 0:
        mout.out("Done.")  # user output


def graphForces(
    trajectory,
    filename=None,
    max=True,
    show=True,
    verbosity=2,
    xlab="Step",
    xmin=None,
    xmax=None,
    ymin=None,
    ymax=None,
    yLog=False,
):
    import mcol
    import mout
    import mplot
    import numpy as np

    if verbosity > 0:
        mout.out(
            "graphing " + mcol.varName + "Forces" + mcol.clear + " ... ",
            printScript=True,
            end="",
        )  # user output

    xdata = []
    fmax = []
    favg = []

    if max:
        ylab = "Force"
    else:
        ylab = "Average Force"

    for n, atoms in enumerate(trajectory):
        xdata.append(n)

        forces = atoms.get_forces()

        this_fmax = np.max(forces)
        this_favg = np.average(forces)

        fmax.append(this_fmax)
        favg.append(this_favg)

    if max:
        ydata = [fmax, favg]
        ytitles = ["Maximum", "Average"]
    else:
        ydata = favg
        ytitles = "Average"

    mplot.graph2D(
        xdata,
        ydata,
        ytitles=ytitles,
        show=show,
        xlab=xlab,
        ylab=ylab,
        filename=filename,
        verbosity=verbosity - 1,
        xmin=xmin,
        xmax=xmax,
        ymin=ymin,
        ymax=ymax,
        ySci=True,
        yLog=yLog,
    )

    if verbosity > 0:
        mout.out("Done.")  # user output


def graphDisplacement(
    trajectory, show=True, filename=None, relative=True, verbosity=2, timestep=None
):
    """
    Root Mean Square Displacement

    """

    import mcol
    import mout
    import mplot
    import numpy as np
    from ase import units

    if verbosity > 0:
        mout.out(
            "graphing " + mcol.varName + "RMSD" + mcol.clear + " ... ",
            printScript=True,
            end="",
        )  # user output

    xdata = []
    rmsd = []

    for n, atoms in enumerate(trajectory):

        if timestep is None:
            xlab = "MD Steps"
            xdata.append(n)
        else:
            xlab = "Time [fs]"
            xdata.append(n * timestep / units.fs)

        positions = atoms.get_positions()

        if relative:
            if n == 0:
                reference = positions.copy()
            positions -= reference

        rmsd.append(np.sqrt(np.mean(positions**2)))

    mplot.graph2D(
        xdata,
        rmsd,
        show=show,
        xlab="MD Steps",
        ylab="RMS Displacement",
        filename=filename,
        verbosity=verbosity - 1,
    )

    if verbosity > 0:
        mout.out("Done.")  # user output


def graphBondLength(
    trajectory,
    indices,
    torsion_indices=None,
    printScript=False,
    show=True,
    filename=None,
    fitMin=None,
    fitMax=None,
    verbosity=2,
    timestep=None,
    title=None,
    fitOrder=None,
    yUnit="Angstroms",
    dataFile=None,
    ymin=0,
    ymax=None,
    xmin=None,
    xmax=None,
    noplot=False,
    write_data=False,
    write_ang=False,
    write_torsion=False,
    write_exang=False,
    debug_plot=False,
    return_data=False,
):
    from .analysis import bondLengthStats

    """
      Graph the bond lengths (displacement) between atoms.

    """

    # import mcol
    import mout
    import mplot
    import numpy as np

    many = any(isinstance(el, list) for el in indices)

    ydata = []
    labels = []

    if timestep is None:
        xlab = "MD Steps"
        xUnit = "MD Step"
    else:
        xlab = "Time [fs]"
        xUnit = "femtoseconds"

    if many:

        for i, pair in enumerate(indices):

            if verbosity > 2:
                print("")
            val, err, label, xdata, this_data = bondLengthStats(
                trajectory,
                pair,
                printScript=printScript,
                verbosity=verbosity - 2,
                timestep=timestep,
                yUnit=yUnit,
                returnData=True,
            )

            labels.append(label)
            ydata.append(this_data)

        if fitOrder is None:
            if not noplot:
                mplot.graph2D(
                    xdata,
                    ydata,
                    ytitles=labels,
                    show=show,
                    xlab=xlab,
                    ylab="Distance [Angstrom]",
                    filename=filename,
                    title=title,
                    verbosity=verbosity - 1,
                    ymin=ymin,
                    ymax=ymax,
                    xmin=xmin,
                    xmax=xmax,
                )
        else:
            if verbosity > 1:
                print("")
            if len(xdata) < fitOrder + 1:
                mout.warningOut("Not enough data-points for fitting.")
                fitOrder = None
            if title is not None:
                dataFile.write("# ")
                dataFile.write(title)
                dataFile.write("\n")
            val, err, fit_func = mplot.fit(
                xdata,
                ydata,
                rank=fitOrder,
                verbosity=verbosity - 1,
                title=title,
                fitMin=fitMin,
                fitMax=fitMax,
                yUnit=yUnit,
                xUnit=xUnit,
                dataFile=dataFile,
            )
            text = mplot.getCoeffStr(val, err, 1, yUnit=yUnit, xUnit=xUnit)
            if not noplot:
                mplot.graph2D(
                    xdata,
                    ydata,
                    fitFunc=fit_func,
                    ytitles=labels,
                    show=show,
                    xlab=xlab,
                    ylab="Distance [Angstrom]",
                    filename=filename,
                    title=title,
                    verbosity=verbosity,
                    subtitle=text,
                    ymin=ymin,
                    ymax=ymax,
                    xmin=xmin,
                    xmax=xmax,
                )

    else:

        if verbosity > 2:
            print("")
        val, err, label, xdata, ydata = bondLengthStats(
            trajectory,
            indices,
            printScript=printScript,
            verbosity=verbosity - 2,
            timestep=timestep,
            yUnit=yUnit,
            returnData=True,
        )

        if fitOrder is None:
            if not noplot:
                mplot.graph2D(
                    xdata,
                    ydata,
                    show=show,
                    xlab=xlab,
                    ylab="Distance [Angstrom]",
                    filename=filename,
                    title=label,
                    verbosity=verbosity - 1,
                    ymin=ymin,
                    ymax=ymax,
                    xmin=xmin,
                    xmax=xmax,
                )
        else:
            if verbosity > 1:
                print("")
            if len(xdata) < fitOrder + 1:
                mout.warningOut("Not enough data-points for fitting.")
                fitOrder = None
            if title is not None:
                dataFile.write("#")
                dataFile.write(title)
                dataFile.write("\n")
            val, err, fit_func = mplot.fit(
                xdata,
                ydata,
                rank=fitOrder,
                verbosity=verbosity - 1,
                fitMin=fitMin,
                fitMax=fitMax,
                title=title,
                yUnit=yUnit,
                xUnit=xUnit,
                dataFile=dataFile,
            )
            text = mplot.getCoeffStr(val, err, 1, yUnit=yUnit, xUnit=xUnit)
            if not noplot:
                mplot.graph2D(
                    xdata,
                    ydata,
                    fitFunc=fit_func,
                    show=show,
                    xlab=xlab,
                    ylab="Distance [Angstrom]",
                    filename=filename,
                    title=label,
                    verbosity=verbosity,
                    subtitle=text,
                    ymin=ymin,
                    ymax=ymax,
                    xmin=xmin,
                    xmax=xmax,
                )

    if write_data:

        import os

        # base=os.path.basename(filename)
        # data_dump = open(os.path.splitext(base)[0]+".dat",'w')
        data_dump = open(filename.replace(".png", ".dat"), "w")

        if many:

            data_dump.write("# x " + str(labels))

            if write_torsion:
                data_dump.write(" opening_angle opening_torsion")
                if write_exang:
                    data_dump.write(" psi phi")
            elif write_ang:
                data_dump.write(" opening_angle")

            data_dump.write("\n")
            data_dump.close()

            # generate the torsion data
            if write_torsion:
                angdata = []
                torsdata = []
                exangdata1 = []
                exangdata2 = []
                for image in trajectory:
                    positions = image.get_positions()
                    symbols = image.get_chemical_symbols()

                    # DC:O2 -> DC:N4
                    # print(1,symbols[torsion_indices[1]],symbols[torsion_indices[0]])
                    vec1 = positions[torsion_indices[1]] - positions[torsion_indices[0]]

                    # DG:N2 -> DG:O6
                    # print(2,symbols[torsion_indices[5]],symbols[torsion_indices[4]])
                    vec2 = positions[torsion_indices[5]] - positions[torsion_indices[4]]

                    # DC:N1 -> DC:N3
                    # print(3,symbols[torsion_indices[3]],symbols[torsion_indices[2]])
                    vec1p = (
                        positions[torsion_indices[3]] - positions[torsion_indices[2]]
                    )

                    # DG:C4 -> DG:N1
                    # print(4,symbols[torsion_indices[7]],symbols[torsion_indices[6]])
                    vec2p = (
                        positions[torsion_indices[7]] - positions[torsion_indices[6]]
                    )

                    vec1pp = np.cross(vec1, vec1p)
                    vec2pp = np.cross(vec2, vec2p)

                    unit_vec1 = vec1 / np.linalg.norm(vec1)
                    unit_vec2 = vec2 / np.linalg.norm(vec2)
                    # unit_vec1p = vec1p / np.linalg.norm(vec1p)
                    # unit_vec2p = vec2p / np.linalg.norm(vec2p)
                    unit_vec1pp = vec1pp / np.linalg.norm(vec1pp)
                    unit_vec2pp = vec2pp / np.linalg.norm(vec2pp)

                    unit_vec1pp_dot_vec2pp = np.dot(unit_vec1pp, unit_vec2pp)
                    tau = np.arccos(unit_vec1pp_dot_vec2pp)
                    torsdata.append(np.pi - tau)

                    unit_vec1_dot_vec2 = np.dot(unit_vec1, unit_vec2)
                    theta = np.arccos(unit_vec1_dot_vec2)

                    # we potentially need to flip the angles
                    vec1_cross_vec2 = np.cross(vec1, vec2)
                    flip = np.dot(vec1_cross_vec2, vec2pp) > 0
                    if flip:
                        theta = -theta
                    angdata.append(theta)

                    if write_exang:
                        # calculate the projected opening angles
                        vec1_dot_vec2 = np.dot(vec1, vec2)
                        vec1ppp = vec1p - np.dot(vec1p, vec1) * vec1
                        vec2ppp = vec2p - np.dot(vec2p, vec2) * vec2
                        psi = np.arctan(np.dot(vec2, vec1ppp) / vec1_dot_vec2)
                        phi = np.arctan(np.dot(vec1, vec2ppp) / vec1_dot_vec2)

                        # Because it is based on atan, it does not need flipping
                        exangdata1.append(psi)
                        exangdata2.append(phi)

                if debug_plot:
                    import matplotlib.pyplot as plt

                    fig, ax = plt.subplots()
                    plt.plot(xdata, ydata[0], label="bot")
                    # plt.plot(xdata,ydata[1],label="mid")
                    plt.plot(xdata, ydata[2], label="top")
                    plt.plot(xdata, angdata, label="ang")
                    plt.plot(xdata, exangdata1, label="psi")
                    plt.plot(xdata, exangdata2, label="phi")
                    plt.plot(xdata, torsdata, label="tors")
                    ax.set_xlim(0, xdata[100])
                    plt.legend(loc="best")
                    plt.show()
                    plt.close()
                    exit()

            # generate the angle data
            elif write_ang:
                angdata = []
                for image in trajectory:
                    positions = image.get_positions()
                    vec1 = positions[indices[0][-1]] - positions[indices[-1][0]]
                    vec2 = positions[indices[0][0]] - positions[indices[-1][-1]]
                    unit_vec1 = vec1 / np.linalg.norm(vec1)
                    unit_vec2 = vec2 / np.linalg.norm(vec2)
                    dot_product = np.dot(unit_vec1, unit_vec2)
                    angle = np.arccos(dot_product)
                    if angle > np.pi / 2:
                        angle = angle - np.pi
                    angdata.append(angle)

            # data_dump = open(os.path.splitext(base)[0]+".dat",'a')
            data_dump = open(filename.replace(".png", ".dat"), "a")

            for index, x in enumerate(xdata):
                # data_dump.write(str(xdata[index])+" ")
                data_dump.write(f"{xdata[index]:.5f} ")
                for data in ydata:
                    # data_dump.write(str(data[index])+" ")
                    data_dump.write(f"{data[index]:.5f} ")
                if write_torsion:
                    # data_dump.write(str(angdata[index])+" ")
                    data_dump.write(f"{angdata[index]:.5f} ")
                    data_dump.write(f"{torsdata[index]:.5f} ")
                    if write_exang:
                        data_dump.write(f"{exangdata1[index]:.5f} ")
                        data_dump.write(f"{exangdata2[index]:.5f} ")
                elif write_ang:
                    # data_dump.write(str(angdata[index])+" ")
                    data_dump.write(f"{angdata[index]:.5f} ")
                data_dump.write("\n")

        else:

            data_dump.write("# x " + label + "\n")
            data_dump.close()

            # data_dump = open(os.path.splitext(base)[0]+".dat",'a')
            data_dump = open(filename.replace(".png", ".dat"), "a")

            for index, x in enumerate(xdata):
                data_dump.write(str(xdata[index]) + " ")
                data_dump.write(str(ydata[index]) + " ")
                data_dump.write("\n")

        data_dump.close()

        mout.out("Data dumped to " + filename.replace(".png", ".dat"))

    if fitOrder is not None:
        return val, err, fit_func
    elif return_data:
        if torsion_indices is not None:
            return xdata, ydata, angdata, torsdata
        return xdata, ydata
    else:
        return None, None, None


def graphBondVibSpec(
    trajectory,
    indices,
    printScript=False,
    show=True,
    filename=None,
    verbosity=2,
    timestep=None,
    title=None,
    ymin=None,
    ymax=None,
    power_spec=False,
    power_segment=None,
    wave_number=False,
    xlab="Frequency [Hz]",
    write_data=False,
):
    # import mcol
    import mout
    import mplot
    from .analysis import getBondLabel

    """
      Graph the fourier transfort of bond lengths (displacement) between atoms.

    """

    # if (verbosity > 0):
    #   mout.out("graphing "+mcol.varName+
    #            title+
    #            mcol.clear+" ... ",
    #            printScript=printScript,
    #            end='') # user output

    many = any(isinstance(el, list) for el in indices)

    if many:
        mout.errorOut("Unsupported.", fatal=True)

    if not many:

        xdata = []
        ydata = []

        bond_label = getBondLabel(trajectory, indices)

        index1 = indices[0]
        index2 = indices[1]

        for n, atoms in enumerate(trajectory):
            # if timestep is None:
            xdata.append(n)

            dist = atoms.get_distance(index1, index2)
            ydata.append(dist)

        # print(len(xdata))
        # print(len(ydata))

        # import scipy
        import scipy.fftpack

        Ydata = scipy.fftpack.fft(ydata)[: len(ydata) // 2]

        if timestep is not None:
            Xdata = scipy.fftpack.fftfreq(n, timestep)[: len(ydata) // 2]
        else:
            Xdata = xdata[: len(ydata) // 2]

        if power_spec:
            if timestep is None:
                mout.errorOut("Timestep required for power spectrum", fatal=True)
            import scipy.signal

            if power_segment is None:
                power_segment = len(ydata) // 10
            f, Pxx_den = scipy.signal.welch(
                ydata, fs=1 / timestep, nperseg=power_segment
            )

            Xdata = f
            Ydata = Pxx_den

        if wave_number:
            temp_x = []
            light_speed = 299792458
            for x in Xdata:
                temp_x.append(x / light_speed / 100.0)
            Xdata = temp_x
            xlab = "Wave Number [/cm]"

        # ydata = np.fft.fft(ydata)

        # if ymin is None:
        #   ymin=min(ydata)
        # if ymax is None:
        #   ymax=max(ydata)

        mplot.graph2D(
            Xdata,
            Ydata,
            ytitles=bond_label,
            show=show,
            filename=filename,
            verbosity=verbosity - 1,
            yLog=True,
            ymin=ymin,
            ymax=ymax,
            xlab=xlab,
        )

        if write_data:

            import os

            base = os.path.basename(filename)
            data_dump = open(os.path.splitext(base)[0] + ".dat", "w")

            if many:

                # NOT UPDATED

                data_dump.write("# x " + str(labels) + "\n")
                data_dump.close()

                data_dump = open(os.path.splitext(base)[0] + ".dat", "a")

                for index, x in enumerate(xdata):
                    data_dump.write(str(xdata[index]) + " ")
                    for data in ydata:
                        data_dump.write(str(data[index]) + " ")
                    data_dump.write("\n")

            else:

                # bond_label = getBondLabel(trajectory,indices)

                if wave_number:
                    data_dump.write("# wave_number [/cm], " + bond_label + "\n")
                else:
                    data_dump.write("# frequency [Hz], " + bond_label + "\n")
                data_dump.close()

                data_dump = open(os.path.splitext(base)[0] + ".dat", "a")

                for index, x in enumerate(Xdata):
                    data_dump.write(str(Xdata[index]) + " ")
                    data_dump.write(str(Ydata[index]) + " ")
                    data_dump.write("\n")

            data_dump.close()


# just a wrapper for mplot.show()
def showFigs(verbosity=1):
    mplot.show(verbosity=verbosity)


def graphBondAngle(
    trajectory,
    indices,
    printScript=False,
    show=True,
    filename=None,
    fitMin=None,
    fitMax=None,
    verbosity=2,
    timestep=None,
    title=None,
    fitOrder=None,
    yUnit="Degrees",
    dataFile=None,
    ymin=0,
    ymax=180,
    xmin=None,
    xmax=None,
):
    import mout
    import mplot
    from .analysis import bondAngleStats

    """
      Graph the bond angle (acute) between atoms.

    """

    many = any(isinstance(el, list) for el in indices)

    ydata = []
    labels = []

    if timestep is None:
        xlab = "MD Steps"
        xUnit = "MD Step"
    else:
        xlab = "Time [fs]"
        xUnit = "picosecond"

    if many:

        for i, triple in enumerate(indices):

            if verbosity > 2:
                print("")
            val, err, label, xdata, this_data = bondAngleStats(
                trajectory,
                triple,
                printScript=printScript,
                verbosity=verbosity - 2,
                timestep=timestep,
                yUnit=yUnit,
                returnData=True,
            )

            labels.append(label)
            ydata.append(this_data)

        if fitOrder is None:
            mplot.graph2D(
                xdata,
                ydata,
                ytitles=labels,
                show=show,
                xlab=xlab,
                ylab="Distance [Angstrom]",
                filename=filename,
                title=title,
                verbosity=verbosity - 1,
            )
        else:
            if verbosity > 1:
                print("")
            if len(xdata) < fitOrder + 1:
                mout.warningOut("Not enough data-points for fitting.")
                fitOrder = None
            if title is not None:
                dataFile.write("#")
                dataFile.write(title)
                dataFile.write("\n")
            val, err, fit_func = mplot.fit(
                xdata,
                ydata,
                rank=fitOrder,
                verbosity=verbosity - 1,
                title=title,
                fitMin=fitMin,
                fitMax=fitMax,
                yUnit=yUnit,
                xUnit=xUnit,
                dataFile=dataFile,
            )
            text = mplot.getCoeffStr(val, err, 1, yUnit=yUnit, xUnit=xUnit)
            mplot.graph2D(
                xdata,
                ydata,
                fitFunc=fit_func,
                ytitles=labels,
                show=show,
                xlab=xlab,
                ylab="Angle [Degrees]",
                filename=filename,
                title=title,
                verbosity=verbosity,
                subtitle=text,
                ymin=ymin,
                ymax=ymax,
                xmin=xmin,
                xmax=xmax,
            )

    else:

        if verbosity > 2:
            print("")
        val, err, label, xdata, ydata = bondAngleStats(
            trajectory,
            indices,
            printScript=printScript,
            verbosity=verbosity - 2,
            timestep=timestep,
            yUnit=yUnit,
            returnData=True,
        )

        if fitOrder is None:
            mplot.graph2D(
                xdata,
                ydata,
                show=show,
                xlab=xlab,
                ylab="Distance [Angstrom]",
                filename=filename,
                title=label,
                verbosity=verbosity - 1,
            )
        else:
            if verbosity > 1:
                print("")
            if len(xdata) < fitOrder + 1:
                mout.warningOut("Not enough data-points for fitting.")
                fitOrder = None
            if title is not None:
                dataFile.write("# ")
                dataFile.write(title)
                dataFile.write("\n")
            val, err, fit_func = mplot.fit(
                xdata,
                ydata,
                rank=fitOrder,
                verbosity=verbosity - 1,
                fitMin=fitMin,
                fitMax=fitMax,
                title=title,
                yUnit=yUnit,
                xUnit=xUnit,
                dataFile=dataFile,
            )
            text = mplot.getCoeffStr(val, err, 1, yUnit=yUnit, xUnit=xUnit)
            mplot.graph2D(
                xdata,
                ydata,
                fitFunc=fit_func,
                show=show,
                xlab=xlab,
                ylab="Distance [Angstrom]",
                filename=filename,
                title=label,
                verbosity=verbosity,
                subtitle=text,
                ymin=ymin,
                ymax=ymax,
                xmin=xmin,
                xmax=xmax,
            )

    if fitOrder is not None:
        return val, err, fit_func
    else:
        return None, None, None



# Site = namedtuple('Site',['types','atoms','position','sidechain','res_name','res_number'])

import numpy as np
import mout
import mcol
from ..residue import Residue
from ..monte import CompoundVolume, CappedCone
import mgo


class Site:
    """Protein Interaction Site"""

    def __init__(
        self,
        interaction_types: list,
        atoms: list,
        position: np.ndarray,
        sidechain: bool,
        res_name: str,
        res_number: int,
        res_chain: str,
    ):

        self.types = interaction_types
        self.atoms = atoms
        self.position = position
        self.sidechain = sidechain
        self.res_name = res_name
        self.res_number = res_number
        self.res_chain = res_chain

        self.atom_numbers = [a.number for a in self.atoms]

        self.accessible = None
        self.in_screen = None

    @property
    def type_str(self):
        return ", ".join(self.types)

    @property
    def atom_names(self):
        return [a.name for a in self.atoms]

    @property
    def atom_str(self):
        return ", ".join(self.atom_names)

    @property
    def name(self):
        return f"{self.res_name} {self.res_number} {self.res_chain} {self.type_str} {self.atom_str}"

    def __repr__(self):
        return self.name

    def summary(self):
        mout.out(f"{mcol.varName}{self.name}{mcol.clear}: {self.position}")

    def accessibility(
        self,
        residue: Residue,
        candidates: list,
        cutoff: float = 5.0,
        proximity: float = 1.5,
        samples: int = 4000,
        skip_hydrogen: bool = True,
        fig=None,
    ):
        """

        residue: parent mp.Residue object of the Site
        candidates: list of mp.Residue objects to consider for occlusion
        cutoff: the radius within which to look for valid binding pocket points
        proximity: minimum radius of binding pocket point
        samples: number of monte carlo samples for volume calculation
        skip_hydrogen: do not calculate occlusion by hydrogen atoms

        """

        self.nearby = residue.get_nearby(candidates, cutoff=cutoff)

        self.forbidden_volume = CompoundVolume(self.position)

        for res in [residue] + self.nearby:
            for atom in res.atoms:
                if skip_hydrogen and atom.species == "H":
                    continue

                d = np.linalg.norm(atom.np_pos - self.position)

                if d < 0.5:
                    continue

                if d > atom.vdw_radius + cutoff:
                    continue

                cone = CappedCone(
                    self.position,
                    atom.np_pos,
                    atom.vdw_radius,
                    name=f"{res.name_number_str} {atom.name_number_str}",
                )

                self.forbidden_volume.add_volume(cone)

        from ..monte import mc_spherical

        if fig is not None:

            fraction, valid_volume, points_in, points_out = mc_spherical(
                self.forbidden_volume,
                cutoff,
                proximity,
                self.position,
                samples,
                boolean=False,
                inverse=True,
                points=True,
            )

            import plotly.graph_objects as go

            trace = mgo.point_trace(points_in, "valid")
            fig.add_trace(trace)
            trace = mgo.point_trace(points_out, "invalid")
            fig.add_trace(trace)

            self.accessible = bool(points_in)

        else:

            self.accessible = mc_spherical(
                self.forbidden_volume,
                cutoff,
                proximity,
                self.position,
                samples,
                boolean=True,
                inverse=True,
            )

        return self.accessible


# add __eq__ to compare to PLIP interaction

# from collections import UserDict

# class SiteList(UserDict):



def compareSystems(system1, system2):
    import mout
    import mcol

    use_cs4qb = True

    try:
        import cs4qb
    except:
        use_cs4qb = False
        mout.warningOut("Not using " + mcol.func + "cs4qb")

    if use_cs4qb:

        import os

        # Get CHARMM36 location
        try:
            c36_path = os.environ["C36CUSTOM"]
        except KeyError:
            mout.errorOut("$C36CUSTOM variable not set", fatal=True)

        # Get FF filepaths
        ffb_filepath = c36_path + "/ffbonded.itp"
        ffnb_filepath = c36_path + "/ffnonbonded.itp"
        rtp_filepath = c36_path + "/merged.rtp"
        hdb_filepath = c36_path + "/merged.hdb"

        # Check if accessible
        for file in (ffb_filepath, rtp_filepath, hdb_filepath):
            if not os.access(file, os.R_OK):
                mout.errorOut(file + " cannot be accessed", fatal=True)

        from cs4qb.gmx_ff import getFFBondList

        system_data = []
        system_data.append({})
        system_data.append({})

        for index, system in enumerate([system1, system2]):

            value_list = []
            bond_list = []
            improper_list = []

            for residue in system.residues:

                this_bond_list, this_improper_list = getFFBondList(
                    residue, rtp_filepath, use_atomtypes=False
                )

                for bond in this_bond_list:

                    if len(bond) == 2:
                        bond_list.append([residue.name + "_" + b for b in bond])
                        value_list.append(residue.get_distance(bond[0], bond[1]))
                    elif len(bond) == 3:
                        bond_list.append([residue.name + "_" + b for b in bond])
                        value_list.append(residue.get_angle(bond[0], bond[1], bond[2]))

                    # else:
                    # value_list.append(None)

            system_data[index]["bond_list"] = bond_list
            system_data[index]["values"] = value_list

        import json

        # Write the CJSON files
        with open("system1.json", "w") as f:
            json.dump(system_data[0], f, indent=4)
        with open("system2.json", "w") as f:
            json.dump(system_data[1], f, indent=4)

        system_data = compare_bond_stat_lists(system_data, dataFile=True)

        with open("compare.json", "w") as f:
            json.dump(system_data[2], f, indent=4)


def compare_bond_stat_lists(system_data, precision=6, diffPrecision=2, dataFile=False):
    import mout
    import mcol

    diff_list = []
    pcnt_diff_list = []

    if dataFile:
        import os

        datfile = "compare.dat"
        f_dat = open(datfile, "w")
        f_dat.write("BOND, DIFFERENCE, % \n")
        f_dat.close()
        f_dat = open(datfile, "a")
    else:
        f_dat = None

    for index, bond in enumerate(system_data[0]["bond_list"]):

        assert system_data[0]["bond_list"][index] == system_data[1]["bond_list"][index]

        if len(bond) == 2:
            bondStr = bond[0] + " - " + bond[1]
        elif len(bond) == 3:
            bondStr = bond[0] + " - " + bond[1] + " - " + bond[2]

        diff, pcnt_diff = mout.differenceOut(
            bondStr,
            system_data[0]["values"][index],
            system_data[1]["values"][index],
            valCol=mcol.result,
            precision=precision,
            diffPrecision=diffPrecision,
            unit="Angstroms",
            dataFile=f_dat,
        )

        diff_list.append(diff)
        pcnt_diff_list.append(pcnt_diff)

    diff_avg = 0
    for diff in pcnt_diff_list:
        diff_avg += diff
    diff_avg /= len(pcnt_diff_list)

    if len(system_data) == 2:
        system_data.append({})
    system_data[2]["bond_list"] = system_data[0]["bond_list"]
    system_data[2]["diff"] = diff_list
    system_data[2]["pcnt_diff"] = pcnt_diff_list
    system_data[2]["avg_diff"] = diff_avg

    f_dat.write("AVERAGE, N/A, " + str(round(diff_avg, 2)) + " \n")

    return system_data

    # print(system1.res_names)
    # print(system2.res_names)

    # print(system1.positions)
    # print(system2.positions)


def euclid_dist(system1, system2):
    import numpy as np

    all_positions1 = []
    all_positions2 = []

    for i, atom in enumerate(system1.atoms):
        all_positions1 += atom.position
        all_positions2 += system2.atoms[i].position

    return np.linalg.norm(np.array(all_positions2) - np.array(all_positions1))



def view(atoms, **kwargs):
    from .system import System

    if isinstance(atoms, System):
        names = [a.name for a in atoms.atoms]
        atoms = atoms.ase_atoms
        for n, a in zip(names, atoms):
            if len(n) > 1:
                try:
                    a.tag = int(n[1:])
                except ValueError:
                    continue
    from ase import visualize

    return visualize.view(atoms, **kwargs)



import mrich


class AtomGroup:
    """General class for a group of atoms. Do not construct this object! Let Molparse handle it"""

    def __init__(self, name: str):

        self._name = name

        # ligand
        self._smiles = None

        # GUI state variables
        self._expand = False
        self._show_context = False
        self._context_options = {}

        self._header_data = []

        from .list import NamedList

        self._atoms = NamedList()

    """ To-Do's:

        - Do not recalculate atomic properties if not necessary

    """

    ### FACTORIES

    @classmethod
    def from_any(cls, name, source):
        """Construct a new AtomGroup from:
        - an mp.NamedList of mp.Atom objects (i.e. mp.System.atoms)
        - an mp.System object
        - an mp.AtomGroup object
        - an mp.Chain object
        - an mp.Residue object
        - a list of ase.Atom objects
        - an ase.Atoms object
        - a list containing any combination of ase.Atom(s), mp.Atom, mp.Chain, mp.Residue, objects
        - a string containing a PDB Block
        """

        import mout

        from ase import Atoms as ase_Atoms
        from ase import Atom as ase_Atom

        from .system import System
        from .chain import Chain
        from .residue import Residue
        from .list import NamedList

        # string (PDB Block)
        if isinstance(source, str):
            return cls.from_pdb_block(source)

        # mp.AtomGroup subclass
        elif issubclass(source.__class__, cls):
            return cls.from_group_subclass(source)

        # mp.AtomGroup
        elif isinstance(source, AtomGroup):
            return cls.from_group(source)

        # NamedList of atoms (i.e. mp.System.atoms):
        elif isinstance(source, NamedList) or isinstance(source, ase_Atoms):
            return cls.from_atoms(name, source)

        # list of various objects
        elif isinstance(source, list):

            from .atom import Atom

            group = cls.__new__(cls)
            group.__init__(name)

            for item in source:

                if issubclass(item.__class__, cls):
                    group = cls.from_group_subclass(item, group=group)

                elif isinstance(item, AtomGroup):
                    group = cls.from_group(item, append=group)

                elif isinstance(item, NamedList) or isinstance(item, ase_Atoms):
                    group = cls.from_atoms(name, item, group=group)

                elif isinstance(item, Atom):
                    group.add_atom(item)

                elif isinstance(item, ase_Atom):
                    group.add_atom(Atom(name=item.symbol, position=item.position))

                else:
                    mout.errorOut(f"Not supported ({item.__class__=})", fatal=True)

            return group

        else:
            mout.errorOut(f"Not supported ({source.__class__=})", fatal=True)

    @classmethod
    def from_pdb_block(cls, pdb_block):

        from .io import parsePDBAtomLine

        name = None
        atom_lines = []

        for line in pdb_block.split("\n"):

            if line.startswith("COMPND"):
                name = line.split()[1]
                continue

            if line.startswith("HETATM") or line.startswith("ATOM"):
                atom_lines.append(line)
                continue

        atoms = []
        for i, line in enumerate(atom_lines):
            atom = parsePDBAtomLine(line, 0, i, 0)
            atoms.append(atom)

        group = cls.from_atoms(name, atoms)

        return group

    @classmethod
    def from_atoms(cls, name, atoms, group=None):

        import mout

        from .list import NamedList
        from ase import Atoms as ase_Atoms

        assert (
            isinstance(atoms, list)
            or isinstance(atoms, NamedList)
            or isinstance(atoms, ase_Atoms)
        )

        # create new object
        if group is None:
            group = cls.__new__(cls)
            group.__init__(name)

        from .atom import Atom
        from .residue import Residue
        from ase import Atom as ase_Atom

        for atom in atoms:

            if isinstance(atom, Atom):
                group.add_atom(atom.copy())

            elif isinstance(atom, ase_Atom):
                group.add_atom(Atom(name=atom.symbol, position=atom.position))

            elif isinstance(atom, Residue):
                for a in atom.atoms:
                    group.add_atom(a.copy())

            else:
                mout.errorOut(
                    "item in named list is neither mp.Atom, ase.Atom, mp.Residue",
                    fatal=True,
                )

        return group

    @classmethod
    def from_group_subclass(cls, source, group=None):

        assert issubclass(source.__class__, cls)

        if group is None:
            group = cls.__new__(cls)
            group.__init__(source.name)

        for atom in source.atoms:
            group.add_atom(atom.copy())

        return group

    @classmethod
    def from_group(cls, group, append=None):
        assert isinstance(group, AtomGroup)
        if append:
            group = group.copy()
            for atom in group.atoms:
                append.add_atom(atom)
            return append

        return group.copy()

    ### PROPERTIES

    @property
    def atoms(self):
        """Child classes of AtomGroup should overload this method"""
        return self._atoms

    @property
    def children(self):
        """Child classes of AtomGroup should overload this method"""
        return self.atoms

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name
        self.fix_names()

    @property
    def num_atoms(self):
        return len(self.atoms)

    @property
    def charge(self):
        """Total charge (float)"""
        charge = 0
        for atom in self.atoms:
            charge += atom.charge
        return charge

    @property
    def atomic_numbers(self):
        """Get all child Atom atomic numbers (list)"""
        return [atom.atomic_number for atom in self.atoms]

    @property
    def positions(self):
        """Get all child Atom positions (list)"""
        return [atom.np_pos for atom in self.atoms]

    @property
    def charges(self):
        """Get all child Atom charges (list)"""
        return [atom.charge for atom in self.atoms]

    @property
    def symbols(self):
        """Get all child Atom symbols (list)"""
        return [atom.symbol for atom in self.atoms]

    @property
    def present_symbols(self):
        return set(self.symbols)

    @property
    def masses(self):
        """Get all child Atom masses (list)"""
        masses = []
        for atom in self.atoms:
            masses.append(atom.mass)
        return masses

    @property
    def species(self):
        """Returns species of all child Atoms (str)"""
        return "".join([atom.species for atom in self.atoms])

    @property
    def atom_indices(self):
        """Returns indices of all child Atoms (list)"""
        return [atom.index for atom in self.atoms]

    @property
    def atom_names(self):
        """Returns indices of all child Atoms (list)"""
        return [atom.name for atom in self.atoms]

    @property
    def atom_numbers(self):
        """Returns indices of all child Atoms (list)"""
        return [atom.number for atom in self.atoms]

    @property
    def bbox(self):
        """Bounding box of the AtomGroup"""
        import numpy as np

        x = [
            min([a.np_pos[0] for a in self.atoms]),
            max([a.np_pos[0] for a in self.atoms]),
        ]
        y = [
            min([a.np_pos[1] for a in self.atoms]),
            max([a.np_pos[1] for a in self.atoms]),
        ]
        z = [
            min([a.np_pos[2] for a in self.atoms]),
            max([a.np_pos[2] for a in self.atoms]),
        ]
        return [x, y, z]

    @property
    def bbox(self):
        """Bounding box of the AtomGroup"""
        import numpy as np

        atoms = self.atoms
        x_coords = [a.np_pos[0] for a in atoms]
        y_coords = [a.np_pos[1] for a in atoms]
        z_coords = [a.np_pos[2] for a in atoms]
        x = [min(x_coords), max(x_coords)]
        y = [min(y_coords), max(y_coords)]
        z = [min(z_coords), max(z_coords)]
        return [x, y, z]

    @property
    def bbox_center(self):
        """Center of the AtomGroups' bounding box"""
        return self._bbox_center()

    @property
    def bbox_sides(self):
        """Bounding box sides of the Atomgroup"""
        return self._bbox_sides()

    @property
    def bbox_norm(self):
        """Length of bounding box diagonal"""
        return self._bbox_norm()

    @property
    def ase_atoms(self):
        """Construct an equivalent ase.Atoms object"""

        from ase import Atoms

        return Atoms(
            symbols=self.symbols, cell=None, pbc=None, positions=self.positions
        )

    @property
    def covalent_radii(self):
        """Get a list of covalent radii"""

        from ase.data import covalent_radii

        radii = []

        for atom in self.atoms:
            radii.append(covalent_radii[atom.atomic_number])

        return radii

    @property
    def contains_alternative_sites(self):
        """Return true if any child atoms have an alternative site defined"""
        return any(a.alternative_site for a in self.atoms)

    @property
    def pdb_block(self):
        from .io import constructPDBAtomLine

        str_buffer = []
        for atom in self.atoms:
            str_buffer.append(constructPDBAtomLine(atom, atom.number, alt_sites=False))
        return "".join(str_buffer)

    @property
    def pdb_block_with_alt_sites(self):
        from .io import constructPDBAtomLine

        str_buffer = []
        for atom in self.atoms:
            str_buffer.append(constructPDBAtomLine(atom, atom.number, alt_sites=True))
        return "".join(str_buffer)

    @property
    def rdkit_mol(self):
        from .rdkit import mol_from_pdb_block

        return mol_from_pdb_block(self.pdb_block)

    @property
    def smiles(self):
        if self._smiles is None:
            from .rdkit import mol_to_smiles

            self._smiles = mol_to_smiles(self.rdkit_mol)
        return self._smiles

    @property
    def vdw_mesh(self):

        import open3d as o3d

        meshes = []

        resolution = 20

        for atom in self.atoms:
            mrich.debug(atom, atom.vdw_radius)

            sphere = o3d.geometry.TriangleMesh.create_sphere(
                radius=atom.vdw_radius, resolution=resolution
            )
            sphere.translate(atom.np_pos)

            meshes.append(sphere)

        merged_mesh = o3d.geometry.TriangleMesh()

        for mesh in meshes:
            merged_mesh += mesh

        merged_mesh.compute_vertex_normals()
        return merged_mesh

    ### INTERNAL METHODS

    def _bbox_center(self, bbox=None):
        import numpy as np

        if bbox is None:
            bbox = self.bbox
        return np.array([np.mean(bbox[0]), np.mean(bbox[1]), np.mean(bbox[2])])

    def _bbox_norm(self, bbox=None):
        import numpy as np

        if bbox is None:
            bbox = self.bbox
        return np.linalg.norm([x[1] - x[0] for x in bbox])

    def _bbox_sides(self, bbox=None):
        if bbox is None:
            bbox = self.bbox
        return [
            bbox[0][1] - bbox[0][0],
            bbox[1][1] - bbox[1][0],
            bbox[2][1] - bbox[2][0],
        ]

    ### METHODS

    def add_atom(self, atom):
        import mout

        atom.parent = self
        if atom.chain is None:
            atom.chain = self._atoms[-1].chain
            mout.warningOut("Taking atom chain from last atom in group!")
        if atom.res_number is None:
            if len(self._atoms):
                atom._res_number = self._atoms[-1].res_number
                mout.warningOut("Taking atom res_number from last atom in group!")
        self._atoms.append(atom)

    def summary(self):
        import mcol
        import mout

        mout.header(
            f"{mcol.varType}AtomGroup{mcol.clear}{mcol.bold}: {mcol.func}{self.name}{mcol.clear}"
        )

        mout.out(
            f'{mcol.underline}Sy name {"index":>6} {"number":>6} {"res":>4} {"res#":>6} {"chain":<5}'
        )

        for a in self.atoms:
            mout.out(f"{mcol.varName}{a.name:<2}{mcol.clear}", end=" ")
            mout.out(f"{mcol.varName}{a.name:<4}{mcol.clear}", end=" ")
            mout.out(f"{a.index:>6}", end=" ")
            mout.out(f"{a.number:>6}", end=" ")
            mout.out(f"{a.residue:>4}", end=" ")
            mout.out(f"{a.res_number:>6}", end=" ")
            mout.out(f"{a.chain:<5}", end="\n")

    def write(self, filename, **kwargs):
        from .io import write

        write(filename, self, **kwargs)

    # overloaded by child classes
    def fix_names(self):
        pass

    def get_atom_by_index(self, index: int, use_pdb_index: bool = True):
        """Get Atom by its index"""
        for atom in self.atoms:
            if use_pdb_index:
                if atom.pdb_index == index:
                    return atom
            else:
                if atom.index == index:
                    return atom

    def centre_of_mass(self, set=None, shift=None):
        """Calculate centre of mass"""
        return self.CoM(set=set, shift=shift)

    def center(self):
        """Move the system's CoM to the origin"""
        return self.CoM(set=[0, 0, 0])

    def CoM(self, set=None, shift=None, verbosity=0):
        """Calculate or manipulate the system's centre of mass.

        if not set and not shift: return CoM
        if set: move the System to the CoM
        if shift: move the System by the specified vector"""

        import mout
        import numpy as np

        position_list = self.positions

        centre_of_mass = np.array(
            [
                sum([pos[0] for pos in position_list]) / len(position_list),
                sum([pos[1] for pos in position_list]) / len(position_list),
                sum([pos[2] for pos in position_list]) / len(position_list),
            ]
        )

        if verbosity > 0:
            mout.varOut(
                "CoM of " + self.name, centre_of_mass, unit="Angstroms", precision=4
            )

        if set is not None:

            assert np.ndim(set) == 1

            try:
                new_com = np.array([set[0], set[1], set[2]])
            except:
                mout.errorOut("Incompatible array input", code="amp.system.CoM.1")

            shift_com = new_com - centre_of_mass

        else:

            shift_com = np.array([0, 0, 0])

        if shift is not None:

            assert np.ndim(shift) == 1

            try:
                new_shift = np.array([shift[0], shift[1], shift[2]])
            except:
                mout.errorOut("Incompatible array input", code="amp.system.CoM.2")

            shift_com = shift_com + new_shift

        if set is not None or shift is not None:

            for atom in self.atoms:
                atom.position = atom.position + shift_com

        return centre_of_mass

    def guess_bonds(self, scale=1.2):
        """Guess connectivity using covalent radii (list of pairs)"""
        import numpy as np
        from ase.gui.view import get_bonds

        ase_atoms = self.ase_atoms
        radii = np.array(self.covalent_radii) * (scale / 1.5)
        bonds = get_bonds(ase_atoms, radii)
        return [[b[0], b[1]] for b in bonds]

    def guess_names(self, target):
        """Try and set the atom names of the system by looking for
        the closest atom of the same species in the target system"""

        import numpy as np

        positions = [b.np_pos for b in target.atoms]
        species = [b.species for b in target.atoms]

        for a in self.atoms:
            a_pos = a.np_pos
            distances = [
                np.linalg.norm(a_pos - p) if s == a.species else 999
                for s, p in zip(species, positions)
            ]
            index = np.argmin(distances)
            b = target.atoms[index]
            a.set_name(b.name, verbosity=0)

    def plot(
        self,
        ax=None,
        color=None,
        center_index=0,
        show=False,
        offset=None,
        padding=1,
        zeroaxis=None,
        frame=False,
        labels=True,
        textdict={"horizontalalignment": "center", "verticalalignment": "center"},
    ):
        """Render the system with matplotlib"""

        import numpy as np
        from ase.visualize.plot import plot_atoms

        # copy the system so as not to alter it
        copy = self.copy()

        # create new axes if none provided
        if ax is None:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots()

        # create colours list
        if color is not None:
            color = [color for a in copy.atoms]

        # create the canvas unit size from bounding box
        canvas = np.array(self.bbox_sides) * 1.5

        # center the render if needed
        if offset is None:
            offset = [0, 0]
            if center_index is not None:
                vec = -copy.atoms[center_index].np_pos + canvas
                copy.CoM(shift=vec, verbosity=0)

        # use ASE to render the atoms
        plot_atoms(
            copy.ase_atoms,
            ax,
            colors=color,
            offset=offset,
            bbox=[0, 0, canvas[0] * 2, canvas[1] * 2],
        )

        # do the labels
        if labels:
            for atom in copy.atoms:
                if atom.species != "H":
                    ax.text(atom.position[0], atom.position[1], atom.name, **textdict)

        # crop the plot
        xmax = copy.bbox[0][1] + padding
        xmin = copy.bbox[0][0] - padding
        ymax = copy.bbox[1][1] + padding
        ymin = copy.bbox[1][0] - padding
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)

        # render the centering axes
        if zeroaxis is not None:
            ax.axvline(canvas[0], color=zeroaxis)
            ax.axhline(canvas[1], color=zeroaxis)

        if not frame:
            ax.axis("off")

        if show:
            plt.show()

        return ax, copy

    def plot3d(
        self,
        extra=[],
        alpha=1.0,
        bonds=True,
        atoms=True,
        velocity=False,
        features=False,
        v_scale=1.0,
        fig=None,
        flat=False,
        show=True,
        transform=None,
        **kwargs,
    ):
        """Render the atoms with plotly graph objects.
        extra can contain pairs of coordinates to be shown as vectors."""

        if bonds:
            bonds = [
                [self.atoms[a].np_pos, self.atoms[b].np_pos]
                for a, b in self.guess_bonds()
            ]
        else:
            bonds = []

        from .amino import AminoAcid

        if isinstance(self, AminoAcid) and isinstance(features, bool) and features:
            features = self.features

        elif not isinstance(features, list) and features:
            from .rdkit import features_from_group

            features = features_from_group(self)

        from .go import plot3d

        return plot3d(
            self.atoms,
            extra,
            bonds,
            alpha,
            plot_atoms=atoms,
            features=features,
            velocity=velocity,
            v_scale=v_scale,
            fig=fig,
            flat=flat,
            show=show,
            transform=transform,
            title=self.name,
            **kwargs,
        )

    def set_coordinates(self, reference, velocity=False):
        """Set all coordinates according to a reference ase.Atoms object"""
        if type(reference) is str:
            if velocity:
                from .io import parse

                sys = parse(reference)
                atoms = sys.atoms
            else:
                from ase.io import read

                atoms = read(reference)
        elif isinstance(reference, list):
            if velocity:
                mout.errorOut("Not supported (group.set_coordinates)", fatal=True)
            for index, atom in enumerate(self.atoms):
                atom.position = reference[index]
            return
        else:
            atoms = reference

        for index, atom in enumerate(self.atoms):
            atom.position = atoms[index].position
            if velocity:
                atom.velocity = atoms[index].velocity

    def rotate(self, angle, vector, center=(0, 0, 0)):
        """Rotate the system (see ase.Atoms.rotate)"""
        ase_atoms = self.ase_atoms
        ase_atoms.rotate(angle, vector, center=center)
        self.set_coordinates(ase_atoms)

    def align_by_posmap(self, map):
        """Align the system to a target by superimposing three shared atoms:

        a --> A
        b --> B
        c --> C

        (a,b,c) are atoms from this system
        (A,B,C) are atoms from the target system

        map should contain Atoms: [[a,b,c],[A,B,C]]

        """

        import numpy as np

        a = map[0][0]
        b = map[0][1]
        c = map[0][2]

        A = map[1][0]
        B = map[1][1]
        C = map[1][2]

        # TRANSLATION

        displacement = A - a
        self.translate(displacement)

        # ROTATION 1

        d = (b + c) / 2
        D = (B + C) / 2
        self.rotate(d - A.np_pos, D - A.np_pos, center=A.np_pos)

        # ROTATION 2

        d = (b + c) / 2
        D = (B + C) / 2

        v_bc = c - b
        v_BC = C - B

        def unit_vector(a):
            return a / np.linalg.norm(a)

        def remove_parallel_component(vector, reference):
            unit_reference = unit_vector(reference)
            projection = np.dot(vector, unit_reference)
            return vector - projection

        v_ad = d - a.np_pos
        v_bc_normal_to_ad = remove_parallel_component(v_bc, v_ad)
        v_BC_normal_to_ad = remove_parallel_component(v_BC, v_ad)
        self.rotate(v_bc_normal_to_ad, v_BC_normal_to_ad, center=A.np_pos)

        # extra stuff for plotly
        extra = [[A.np_pos, A.np_pos + v_ad]]
        return extra

    def copy(self):
        import copy

        return copy.deepcopy(self)

    def get_nearby(self, candidates, cutoff):
        """Get a subset of <candidates> that have at least one atom within <cutoff> of any atom in this AtomGroup."""

        import numpy as np

        # precompute properties of self
        self_atoms = self.atoms
        self_bbox = self.bbox
        self_bbox_halfnorm = self._bbox_norm(self_bbox) / 2
        self_bbox_center = self._bbox_center(self_bbox)

        def is_intersecting():

            for self_atom in self_atoms:

                d = np.linalg.norm(self_atom.np_pos - group_bbox_center)

                # skip if atom is not within bbox_halfnorm + cutoff of the group
                if d > cutoff + group_bbox_halfnorm:
                    continue

                self_atom_np_pos = self_atom.np_pos

                for group_atom in group.atoms:

                    d = np.linalg.norm(self_atom_np_pos - group_atom.np_pos)

                    # actually test for inter-atomic distance
                    if d <= cutoff:
                        return True

        # go through list of candidates
        nearby = []
        for group in candidates:

            # precompute properties of self
            group_bbox = group.bbox
            group_bbox_halfnorm = group._bbox_norm(group_bbox) / 2
            group_bbox_center = group._bbox_center(group_bbox)

            center_delta = np.linalg.norm(self_bbox_center - group_bbox_center)
            if center_delta == 0.0:
                continue
            min_bbox_separation = (
                center_delta - self_bbox_halfnorm - group_bbox_halfnorm
            )

            # discard any candidate where the boundingboxes are definitely more than cutoff apart
            if min_bbox_separation > cutoff:
                continue

            # check for intersection
            intersects = is_intersecting()

            if intersects:
                nearby.append(group)

        return nearby

    ### GUI THINGS

    # open GUI tree viewer
    def tree(self):
        from .tree import tree

        tree(self)

    # expand GUI tree view
    def expand(self):
        self._expand = True
        for child in self.children:
            child._expand = False

    # collapse GUI tree view
    def collapse(self):
        self._expand = False

    # open GUI context window
    def spawn_context(self, line, col):
        self._tree.log(f"trying to spawn context window at {line} {col}")
        self._show_context = True
        self._tree.spawn_context(self, line, col)

    # close GUI context window
    def hide_context(self):
        self._tree.log(f"trying to hide context window")
        self._show_context = False
        self._tree.hide_context_menu()

    @property
    def _context_info(self):

        com = self.CoM(verbosity=0)
        com = f"[{com[0]:.2f} {com[1]:.2f} {com[2]:.2f}]"

        bbox = self.bbox
        bbox = f"[[{bbox[0][0]:.2f} {bbox[0][1]:.2f}] [{bbox[1][0]:.2f} {bbox[1][1]:.2f}] [{bbox[2][0]:.2f} {bbox[2][1]:.2f}]]"

        bbox_sides = self.bbox_sides
        bbox_sides = f"[{bbox_sides[0]:.2f} {bbox_sides[1]:.2f} {bbox_sides[2]:.2f}]"

        items = {
            "Centre of Mass": com,
            "Bounding Box": bbox,
            "Bounding Box Sides": bbox_sides,
            "Bounding Box Diagonal": f"{self.bbox_norm:.2f}",
            "Total Charge": self.charge,
            "Total #Atoms": self.num_atoms,
        }

        from .system import System
        from .group import AtomGroup

        if not isinstance(self, System) and not isinstance(self, AtomGroup):
            file_out = f"{self.__class__.__name__}_{self.name}_{self.index}.pdb"

            clickables = {
                f'Export PDB "{file_out}"': lambda x: x.write(file_out),
            }

        else:
            clickables = {}

        if self.num_atoms == 1:
            items.pop("Bounding Box")
            items.pop("Bounding Box Sides")
            items.pop("Bounding Box Diagonal")

        return items, clickables

    ### DUNDERS

    def __repr__(self):
        if not self.name:
            return "AtomGroup"
        return self.name

    def __str__(self):
        if not self.name:
            return "AtomGroup"
        return self.name

    def __len__(self):
        return len(self.children)

    def __getitem__(self, key):
        """Access child Atom, Residue or Chain by index or name

        group['a0'] : returns the first atom
        group['an10'] : returns the residue with number/resid 10
        group['aH1'] : returns all atoms named 'H1'

        group['r0'] : returns the first residue
        group['rn10'] : returns the residue with number/resid 10
        group['rASN'] : returns all residues named 'ASN'

        group['c0'] : returns the first chain
        group['cA'] : returns all chains named 'A'
        """

        if not isinstance(key, str):
            import mout

            mout.error(
                f"Cannot index an AtomGroup with a non-string {key=}. {self=}. See help(molparse.AtomGroup.__getitem__)"
            )
            raise IndexError(f"mp.AtomGroup.__getitem__ received non-string key")

        try:

            if key[0] == "a":
                return self.atoms[key[1:]]

            elif key[0] == "r":
                return self.residues[key[1:]]

            elif key[0] == "c":
                return self.chains[key[1:]]

            else:
                import mout

                mout.errorOut(
                    "key must start with 'a', 'r', or 'c' to refer to an Atom, Residue, or Chain, respectively. See help(molparse.group.__getitem__)"
                )

        except AttributeError:

            import mout

            lookup = {"a": "atoms", "c": "chains", "r": "residues"}
            mout.errorOut(f"{type(self)} does not possess {lookup[key[0]]} attribute")


# def __del__(self):
# 	import mout
# 	mout.debug(f'{self} was deleted')



import mrich as logger


def sdf_align(
    sdf_path: str,
    source_protein_path: str,
    target_protein_path: str,
    mol_col: str = "ROMol",
    name_col: str = "ID",
    backbone_only: bool = True,
    heavy_atoms_only: bool = True,
):
    """Align molecules in an SDF

    :param sdf_path: input file
    :param source_protein_path: source protein conformation
    :param target_protein_path: target protein conformation
    :param mol_col: name of molecule object column
    :param name_col: name of molecule ID column
    :param backbone_only: align using protein backbone atoms only
    :param heavy_atoms_only: align using protein heavy atoms only

    """

    from ..io import parse
    from numpy import array, double
    from rdkit.Chem import PandasTools, rdMolTransforms, Mol
    from .mol import mol_from_smiles, mol_to_smiles, mol_to_pdb_block

    # parse the reference files
    source = parse(source_protein_path)
    target = parse(target_protein_path)

    # align the proteins
    source_center, target_center, rotation = source.align_to(
        target,
        protein_only=True,
        return_transformations=True,
        backbone_only=backbone_only,
        heavy_atoms_only=heavy_atoms_only,
    )

    # create the transformation matrices

    translation_1 = array(
        [
            [1, 0, 0, -source_center[0]],
            [0, 1, 0, -source_center[1]],
            [0, 0, 1, -source_center[2]],
            [0, 0, 0, 1],
        ],
        dtype=double,
    )

    translation_2 = array(
        [
            [1, 0, 0, target_center[0]],
            [0, 1, 0, target_center[1]],
            [0, 0, 1, target_center[2]],
            [0, 0, 0, 1],
        ],
        dtype=double,
    )

    rotation = array(
        [
            [*rotation[0], 0],
            [*rotation[1], 0],
            [*rotation[2], 0],
            [0, 0, 0, 1],
        ],
        dtype=double,
    )

    df = PandasTools.LoadSDF(sdf_path)

    for i, row in df.iterrows():
        if row[name_col] == "ver_1.2":
            logger.warning("Skipping Fragalysis header molecule")
            continue

        mol = row[mol_col]

        # translate to origin
        rdMolTransforms.TransformConformer(mol.GetConformer(0), translation_1)

        # apply rotation
        rdMolTransforms.TransformConformer(mol.GetConformer(0), rotation)

        # translate to target
        rdMolTransforms.TransformConformer(mol.GetConformer(0), translation_2)

    out_path = sdf_path.replace(".sdf", "_aligned.sdf")

    logger.writing(out_path)
    PandasTools.WriteSDF(df, out_path, mol_col, name_col, list(df.columns))


def sdf_combine(files: list, output: str, mol_col: str = "ROMol", name_col: str = "ID"):
    """Concatenate SDF files

    :param files: generator or iterable of file names/paths
    :param output: file name/path of output SDF
    :param mol_col: name of molecule object column
    :param name_col: name of molecule ID column

    """

    from rdkit.Chem import PandasTools
    from pathlib import Path
    from pandas import concat
    import mrich

    out_path = Path(output)

    dfs = []
    for file in mrich.track(files, prefix="parsing SDFs"):
        in_path = Path(file)

        df = PandasTools.LoadSDF(in_path)

        dfs.append(df)

    df = concat(dfs, ignore_index=True)

    mrich.writing(out_path)
    PandasTools.WriteSDF(
        df, out_path, molColName=mol_col, idName=name_col, properties=df.columns
    )



def SPE(
    control="CONTROL",
    config="CONFIG",
    field="FIELD",
    workdir="DL_POLY",
    num_procs=8,
    log="_amp.dl_poly.log",
    verbosity=2,
    output="OUTPUT",
):
    import os
    import mout
    import mcol

    if verbosity > 0:
        mout.headerOut(
            "Calling " + mcol.func + "DL_POLY" + mcol.clear + mcol.bold + "..."
        )

    if verbosity > 1:
        mout.varOut("IO: Workdir", workdir, valCol=mcol.file)
        mout.varOut("IO: Field", field, valCol=mcol.file)
        mout.varOut("IO: Config", config, valCol=mcol.file)
        mout.varOut("IO: Control", control, valCol=mcol.file)
        mout.varOut("IO: Output", output, valCol=mcol.file)

    if verbosity < 2:
        mout.redirectPrint(log)

    from dlpoly import DLPoly

    dl_poly_root = os.environ["DL_POLY"]
    dlp_exec = dl_poly_root + "/execute/DLPOLY.Z"

    DL_POLY = DLPoly(
        control=control, config=config, field=field, workdir=workdir, output=output
    )

    DL_POLY.run(executable=dlp_exec, numProcs=num_procs)

    DL_POLY.load_statis(workdir + "/STATIS")

    if verbosity < 2:
        mout.enablePrint()

    # Parse DL_POLY's OUTPUT
    num_warn = int(
        os.popen("grep 'warning - ' " + workdir + "/" + output + " | wc -l").read()
    )
    num_errs = int(
        os.popen("grep 'error - ' " + workdir + "/" + output + " | wc -l").read()
    )

    if num_warn > 0:
        mout.warningOut(
            "DL_POLY issued " + str(num_warn) + " warnings.",
            code="amp.dl_poly.dl_poly_py.SPE[1]",
        )
    if num_errs > 0:
        mout.errorOut(
            "DL_POLY issued " + str(num_errs) + " errors.",
            fatal=True,
            code="amp.dl_poly.dl_poly_py.SPE[2]",
        )

    e_tot = totalEnergy(DL_POLY, verbosity=verbosity - 1)

    return e_tot


def totalEnergy(dlpoly, verbosity=1):
    import mout
    import mcol

    if dlpoly.statis is None:
        mout.errorOut("STATIS file not loaded", code="amp.dl_poly.totalEnergy[1]")
        return None

    e_tot = dlpoly.statis.data[:, 3][0]

    if verbosity > 0:
        mout.varOut("Total Energy", e_tot, unit="kcal/mol", valCol=mcol.result)

    return e_tot



def plot3d(
    atoms,
    extra=[],
    bonds=[],
    alpha=1.0,
    velocity=False,
    v_scale=1.0,
    fig=None,
    flat=False,
    show=True,
    transform=None,
    title=None,
    plot_atoms=True,
    features=None,
    extra_labels=None,
):
    """Render the atoms with plotly graph objects.
    extra can contain pairs of coordinates to be shown as vectors."""

    import plotly.graph_objects as go
    from ase.data import vdw_radii, atomic_numbers
    from ase.data.colors import jmol_colors

    from .group import AtomGroup

    import mout

    group = AtomGroup.from_any("Compound AtomGroup", atoms)
    atoms = group.atoms

    species = [a.symbol for a in atoms]
    species = list(set(species))

    if fig is None:
        fig = go.Figure()

    if features:

        from .rdkit import Feature

        if isinstance(features[0], Feature):
            features = [
                dict(
                    family=f.family,
                    position=f.position,
                    indices=None,
                    x=f.x,
                    y=f.y,
                    z=f.z,
                )
                for f in features
            ]

        import plotly.express as px

        px_fig = px.scatter_3d(
            features,
            x="x",
            y="y",
            z="z",
            color="family",
            size_max=100,
            size=[100] * len(features),
            opacity=0.5,
        )

        for trace in px_fig.data:
            fig.add_trace(trace)

    if bonds:
        x, y, z = [], [], []

        for a, b in bonds:
            x.extend([a[0], b[0], None])
            y.extend([a[1], b[1], None])
            z.extend([a[2], b[2], None])

        if transform and flat:
            x, y = zip(*transform(list(zip(x, y))))

        if flat:
            trace = go.Scatter(
                x=x, y=y, mode="lines", name="bonds", line=dict(color="black", width=16)
            )
        else:
            trace = go.Scatter3d(
                x=x,
                y=y,
                z=z,
                mode="lines",
                name="bonds",
                line=dict(color="black", width=16),
            )

        fig.add_trace(trace)

    if not flat and velocity:

        x, y, z = [], [], []

        for atom in atoms:

            if atom.velocity is None:
                mout.warningOut(f"Atom {atom.name} (#{atom.number}) has no velocity")
                continue

            a = atom.np_pos
            b = v_scale * atom.np_vel + a

            x.extend([a[0], b[0], None])
            y.extend([a[1], b[1], None])
            z.extend([a[2], b[2], None])

        trace = go.Scatter3d(
            x=x,
            y=y,
            z=z,
            mode="lines",
            name="velocity",
            line=dict(color="red", width=16),
        )

        fig.add_trace(trace)

    if plot_atoms:
        for s in species:

            atom_subset = [a for a in atoms if a.symbol == s]

            data = {}

            positions = [a.np_pos for a in atom_subset]
            x, y, z = zip(*positions)

            if transform and flat:
                x, y = zip(*transform(list(zip(x, y))))

            data["x"] = x
            data["y"] = y
            data["z"] = z

            atomic_number = atomic_numbers[s]
            size = vdw_radii[atomic_number] * 15
            color = jmol_colors[atomic_number]
            color = (color[0] * 256, color[1] * 256, color[2] * 256, alpha)

            data["index"] = [a.index for a in atom_subset]
            data["residue"] = [a.residue for a in atom_subset]
            data["res_number"] = [a.res_number for a in atom_subset]
            data["name"] = [a.name for a in atom_subset]

            customdata = [
                f"name={a.name}<br>chain={a.chain}<br>index={a.index}<br>number={a.number}<br>residue={a.residue}<br>res_index={a.res_index}<br>res_number={a.res_number}<br>x={a.x:.3f}<br>y={a.y:.3f}<br>z={a.z:.3f}"
                for a in atom_subset
            ]

            if flat:
                trace = go.Scatter(
                    x=x,
                    y=y,
                    mode="markers",
                    name=s,
                    marker=dict(
                        size=size,
                        color=f"rgba{color}",
                        line=dict(color="black", width=2),
                    ),
                    customdata=customdata,
                    hovertemplate="%{customdata}<extra></extra>",
                )
            else:
                trace = go.Scatter3d(
                    x=x,
                    y=y,
                    z=z,
                    mode="markers",
                    name=s,
                    marker=dict(
                        size=size,
                        color=f"rgba({color[0]}, {color[1]}, {color[2]})",
                        line=dict(color="black", width=2),
                    ),
                    customdata=customdata,
                    hovertemplate="%{customdata}<extra></extra>",
                )

            fig.add_trace(trace)

    if not flat:

        if extra_labels:

            for (a, b), l in zip(extra, extra_labels):
                trace = go.Scatter3d(
                    x=[a[0], b[0]], y=[a[1], b[1]], z=[a[2], b[2]], name=l
                )

                fig.add_trace(trace)

        else:
            for i, (a, b) in enumerate(extra):
                trace = go.Scatter3d(
                    x=[a[0], b[0]], y=[a[1], b[1]], z=[a[2], b[2]], name=f"extra[{i}]"
                )

                fig.add_trace(trace)

    fig.update_layout(scene_aspectmode="data")

    if title:
        fig.update_layout(title=title)

    if show:
        fig.show()

    return fig



import mout
from rdkit import Chem, Geometry


# class Molecule(Chem.rdchem.Mol):
#   def __init__(self,*args,**kwargs):
#       super(ClassName, self).__init__(*args,**kwargs)


def mol_from_pdb_block(pdb_block):
    return Chem.rdmolfiles.MolFromPDBBlock(pdb_block, removeHs=False)


def mol_to_smiles(mol):
    return Chem.MolToSmiles(mol)


def mol_from_smiles(smiles):
    return Chem.MolFromSmiles(smiles)


def mol_to_pdb_block(mol, hetatm_only: bool = False):
    if hetatm_only:
        buffer = Chem.MolToPDBBlock(mol)

        new = []
        for line in buffer.split("\n"):
            if line.startswith("HETATM"):
                new.append(line)
        return "\n".join(new)

    else:
        return Chem.MolToPDBBlock(mol)


def mol_to_AtomGroup(mol):
    from ..group import AtomGroup

    group = AtomGroup.from_pdb_block(mol_to_pdb_block(mol))
    if hasattr(mol, "_Name"):
        group.name = mol._Name
    return group


def mol_from_AtomGroup(group):
    return group.rdkit_mol


def protonate(mol, embed=True, align=True, verbosity=1):
    mol_prot = Chem.AddHs(mol)
    if embed and align:
        if verbosity:
            mout.warning("May lose exact pose", code="mp.rdkit.protonate.1")
        mol = Chem.AllChem.ConstrainedEmbed(mol_prot, mol)
    elif embed:
        if verbosity:
            mout.warning(
                "Disregarding original coordinates", code="mp.rdkit.protonate.2"
            )
        ps = Chem.AllChem.ETKDGv3()
        Chem.AllChem.EmbedMolecule(mol_prot, ps)
    return mol_prot


def copy_mol(mol):
    return Chem.Mol(mol)


def mol_from_cif(cif_file):

    from gemmi import cif
    from .xca_utils import strip_quotes, BOND_TYPES

    mol = Chem.RWMol()
    conf = Chem.Conformer()

    doc = cif.read(str(cif_file))

    # Diamond CIFs have two blocks, but the one we want will be named data_comp_LIG
    block = doc.find_block("comp_LIG")

    # Other CIFs have unpredictable block names, so let's hope there is only one
    if not block:
        block = doc.sole_block()
    if not block:
        print("sole block not found")
        return None

    comp_ids = block.find_loop("_chem_comp_atom.comp_id")
    atom_ids = block.find_loop("_chem_comp_atom.atom_id")
    atom_symbols = block.find_loop("_chem_comp_atom.type_symbol")
    # coordinates are sometimes called "x" and sometimes "model_Cartn_x" etc.
    x = block.find_loop("_chem_comp_atom.x")
    if not x:
        x = block.find_loop("_chem_comp_atom.model_Cartn_x")
    y = block.find_loop("_chem_comp_atom.y")
    if not y:
        y = block.find_loop("_chem_comp_atom.model_Cartn_y")
    z = block.find_loop("_chem_comp_atom.z")
    if not z:
        z = block.find_loop("_chem_comp_atom.model_Cartn_z")
    charges = [0] * len(atom_ids)
    if block.find_loop("_chem_comp_atom.charge"):
        charges = list(block.find_loop("_chem_comp_atom.charge"))
    elif block.find_loop("_chem_comp_atom.partial_charge"):
        charges = list(block.find_loop("_chem_comp_atom.partial_charge"))

    atoms = {}
    ligand_name = None
    for name, s, id, px, py, pz, charge in zip(
        comp_ids, atom_symbols, atom_ids, x, y, z, charges
    ):
        # sometimes that atom ids are wrapped in double quotes
        if ligand_name is None:
            ligand_name = name
        elif name != ligand_name:
            print(
                "WARNING: ligand name has changed from {} to {}. Old name will be used.".format(
                    ligand_name, name
                )
            )

        id = strip_quotes(id)

        if len(s) == 2:
            s = s[0] + s[1].lower()

        atom = Chem.Atom(s)
        atom.SetFormalCharge(round(float(charge)))
        atom.SetProp("atom_id", id)
        idx = mol.AddAtom(atom)
        atom.SetIntProp("idx", idx)
        atoms[id] = atom

        point = Geometry.Point3D(float(px), float(py), float(pz))
        conf.SetAtomPosition(idx, point)

    atom1 = block.find_loop("_chem_comp_bond.atom_id_1")
    atom2 = block.find_loop("_chem_comp_bond.atom_id_2")
    bond_type = block.find_loop("_chem_comp_bond.type")
    if not bond_type:
        bond_type = block.find_loop("_chem_comp_bond.value_order")

    for a1, a2, bt in zip(atom1, atom2, bond_type):
        mol.AddBond(
            atoms[strip_quotes(a1)].GetIntProp("idx"),
            atoms[strip_quotes(a2)].GetIntProp("idx"),
            BOND_TYPES[bt],
        )

    Chem.SanitizeMol(mol)
    mol.AddConformer(conf)
    Chem.AssignStereochemistryFrom3D(mol)
    mol = Chem.RemoveAllHs(mol)

    mol.SetProp("_Name", ligand_name)

    return mol



from ase.calculators.general import Calculator


class DL_POLY(Calculator):
    import mout
    import mcol
    from ase.calculators.calculator import compare_atoms, PropertyNotImplementedError
    import os

    def __init__(
        self,
        field,
        control,
        workdir="DL_POLY",
        name="DL_POLY",
        output="OUTPUT",
        verbosity=1,
        debug=False,
        num_procs=8,
        backup=False,
    ):

        if debug:
            mout.debugOut(name + ".__init__")

        # Manage arguments
        self.name = name
        self.workdir = workdir
        self.field = field
        self.control = control
        self.output = output
        self.debug = debug
        if debug:
            verbosity = 99
        self.verbosity = verbosity
        self.energy_zero = None
        self.energy_free = None
        self.forces = None
        self.old_atoms = None
        self.run_count = 0
        self.num_procs = num_procs
        self.backup = backup

        self.field_path = self.workdir + "/" + self.field
        self.control_path = self.workdir + "/" + self.control
        self.output_path = self.workdir + "/" + self.output
        self.revcon_path = self.workdir + "/REVCON"
        self.history_path = self.workdir + "/HISTORY"
        self.statis_path = self.workdir + "/STATIS"
        self.auto_config = "auto.CONFIG"
        self.auto_config_path = self.workdir + "/" + self.auto_config

        self.results = {"energy": self.energy_zero, "forces": self.forces}

        self.implemented_properties = ["energy", "forces"]

        if self.verbosity > 1:
            mout.varOut("IO: Workdir", self.workdir, valCol=mcol.file)
            mout.varOut("IO: Field", self.field, valCol=mcol.file)
            mout.varOut("IO: Control", self.control, valCol=mcol.file)
            mout.varOut("IO: Output", self.output, valCol=mcol.file)

        self.clean_directory(full=True, verbosity=verbosity - 2)

    def update(self, atoms):
        if self.debug:
            mout.debugOut(self.name + ".update")

        # Check if an update is required:

        if self.old_atoms is None:  # No calculation has been run yet

            if self.verbosity > 1:
                mout.out("Calculation needed (first run)")

            self.calculate(atoms)

        else:  # Previous calculation data exists

            # Get the names of changed properties
            changed_properties = compare_atoms(atoms, self.old_atoms)

            # If there are changes, run DL_POLY SPE
            if len(changed_properties) > 0:
                if self.verbosity > 1:
                    mout.out("Calculation needed (atoms have changed)")
                    mout.varOut("Changed Properties", changed_properties)
                self.calculate(atoms)
            else:
                if self.verbosity > 1:
                    mout.out("No calculation needed (atoms are unchanged)")

    def clean_directory(self, full=False, verbosity=1):
        if self.debug:
            mout.debugOut(self.name + ".clean_directory")

        if full:  # Remove the work directory
            if verbosity > 0:
                os.system("rm -rfv " + self.workdir)
            else:
                os.system("rm -rf " + self.workdir)
        else:
            if verbosity > 0:  # Only remove output files
                os.system("rm -f " + self.revcon_path)
                os.system("rm -f " + self.output_path)
                os.system("rm -f " + self.history_path)
                os.system("rm -f " + self.statis_path)
            else:
                os.system("rm -fv " + self.revcon_path)
                os.system("rm -fv " + self.output_path)
                os.system("rm -fv " + self.history_path)
                os.system("rm -fv " + self.statis_path)

    def calculate(self, atoms):
        from molparse import write, read
        from .dl_poly_py import SPE

        if self.debug:
            mout.debugOut(self.name + ".calculate")

        # Check pre-requisites
        if self.field is None:
            mout.errorOut("No FIELD specified")
        if self.control is None:
            mout.errorOut("No CONTROL specified")

        if self.run_count > 0:
            if self.backup:
                os.system(
                    "cp "
                    + self.auto_config_path
                    + " "
                    + self.workdir
                    + "/backup"
                    + str(self.run_count)
                    + ".CONFIG"
                )
        else:
            os.system("mkdir -p " + self.workdir)

        # Write new system configuration in DL_POLY format
        write(self.auto_config, atoms, format="dlp4", verbosity=self.verbosity - 1)

        # Get rid of old outputs
        self.clean_directory(full=False, verbosity=self.verbosity - 2)

        # Call DL_POLY
        e_tot = SPE(
            control=self.control,
            config=self.auto_config,
            field=self.field,
            workdir=self.workdir,
            output=self.output,
            num_procs=self.num_procs,
            verbosity=self.verbosity - 1,
        )

        # Clean up files in root directory
        os.system("rm " + self.auto_config)
        os.system("rm OUTPUT")

        # Store the atoms on which the calculation was run
        self.old_atoms = atoms.copy()

        # Get the trajectory
        history = read(
            self.history_path, format="dlp-history", verbosity=self.verbosity - 1
        )

        # Store the calculation results
        self.energy_zero = e_tot
        self.energy_free = e_tot
        self.forces = history.get_forces()
        self.results = {"energy": self.energy_zero, "forces": self.forces}

        # Iterate the run_count
        self.run_count += 1

    def get_forces(self, atoms):
        if self.debug:
            mout.debugOut(self.name + ".get_forces")
        self.update(atoms)
        return self.forces

    def get_stress(self, atoms):
        raise PropertyNotImplementedError

    def get_dipole(self, atoms):
        raise PropertyNotImplementedError

    def get_charges(self, atoms):
        raise PropertyNotImplementedError

    def get_magmom(self, atoms):
        raise PropertyNotImplementedError

    def get_magmoms(self, atoms):
        raise PropertyNotImplementedError

    def reset(self):
        self.energy_zero = None
        self.forces = None
        self.old_atoms = None
        self.energy_free = None
        self.run_count = 0
        self.results = {"energy": self.energy_zero, "forces": self.forces}

    # def set_field(self,field):
    #   if self.debug: mout.debugOut(self.name+".set_field")
    #   self.field = field

    # def set_control(self,control):
    #   if self.debug: mout.debugOut(self.name+".set_control")
    #   self.control = control
